#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC, abstractmethod
from collections import OrderedDict
import numpy as np
from typing import Tuple, List, Union

from nomad.units import ureg
from nomad.metainfo import Section
from nomad.utils import RestrictedDict
from nomad import config
from nomad.datamodel.results import (
    Method,
    Electronic,
    Simulation,
    DFT,
    GW,
    xc_treatments,
)


class MethodNormalizer():
    def __init__(self, entry_archive, repr_system, material, logger):
        self.entry_archive = entry_archive
        self.repr_system = repr_system
        self.repr_method = None
        self.material = material
        self.method_name = None
        self.run = self.entry_archive.run[0]
        self.logger = logger

    def method(self) -> Method:
        """Returns a filled Method section.

        Returns:
            Filled method section.
        """
        method = Method()
        simulation = Simulation()
        repr_method = None
        method_name = config.services.unavailable_value
        methods = self.run.method
        n_methods = len(methods)

        def get_method_name(section_method):
            method_name = config.services.unavailable_value
            if section_method.electronic and section_method.electronic.method:
                method_name = section_method.electronic.method
            else:
                if section_method.gw is not None:
                    method_name = "GW"
            return method_name

        # If only one method is specified, use it directly
        if n_methods == 1:
            repr_method = methods[0]
            method_name = get_method_name(repr_method)
        # If several methods have been declared, we need to find the "topmost"
        # method and report it.
        elif n_methods > 1:
            # Method referencing another as "core_method". If core method was
            # given, create new merged method containing all the information.
            for sec_method in methods:
                core_method = sec_method.core_method_ref
                if core_method is not None:
                    if sec_method.electronic:
                        electronic = core_method.electronic
                        electronic = electronic if electronic else core_method.m_create(Electronic)
                        core_method.electronic.method = sec_method.electronic.method
                    repr_method = core_method
                    method_name = get_method_name(repr_method)

            # Perturbative methods: we report the "topmost" method (=method
            # that is referencing something but is not itself being
            # referenced).
            referenced_methods = set()
            for sec_method in methods:
                starting_method_ref = sec_method.starting_method_ref
                if starting_method_ref is not None:
                    referenced_methods.add(starting_method_ref.m_path())
            if len(referenced_methods) == n_methods - 1:
                for sec_method in methods:
                    if sec_method.m_path() not in referenced_methods:
                        method_name = get_method_name(sec_method)
                        if method_name != config.services.unavailable_value:
                            repr_method = sec_method
                            break
        self.repr_method = repr_method
        self.method_name = method_name
        settings_basis_set = get_basis_set(self.entry_archive, self.repr_method, self.repr_system, self.logger)
        functional_long_name = self.functional_long_name()
        method.workflow_name = self.workflow_name()

        if method_name == "GW":
            method.method_name = "GW"
            gw = GW()
            gw.type = repr_method.gw.type
            gw.starting_point = repr_method.gw.starting_point.split()
            simulation.gw = gw
        elif method_name in {"DFT", "DFT+U"}:
            method.method_name = "DFT"
            dft = DFT()
            dft.basis_set_type = self.basis_set_type()
            dft.basis_set_name = self.basis_set_name()
            method.method_id = self.method_id_dft(settings_basis_set, functional_long_name)
            method.parameter_variation_id = self.parameter_variation_id_dft(settings_basis_set, functional_long_name)
            dft.core_electron_treatment = self.core_electron_treatment()
            if repr_method.electronic is not None:
                if repr_method.electronic.smearing is not None:
                    dft.smearing_kind = repr_method.electronic.smearing.kind
                    dft.smearing_width = repr_method.electronic.smearing.width
                if repr_method.electronic.n_spin_channels:
                    dft.spin_polarized = repr_method.electronic.n_spin_channels > 1
                dft.van_der_Waals_method = repr_method.electronic.van_der_waals_method
                dft.relativity_method = repr_method.electronic.relativity_method
            dft.xc_functional_names = self.xc_functional_names()
            dft.xc_functional_type = self.xc_functional_type(dft.xc_functional_names)
            if repr_method.scf is not None:
                dft.scf_threshold_energy_change = repr_method.scf.threshold_energy_change
            simulation.dft = dft

        method.equation_of_state_id = self.equation_of_state_id(method.method_id, self.material.chemical_formula_hill)
        simulation.program_name = self.run.program.name
        simulation.program_version = self.run.program.version
        method.simulation = simulation
        return method

    def workflow_name(self):
        workflow_name = None
        if self.entry_archive.workflow:
            workflow_name = set(filter(lambda x: x, map(lambda x: x.type, self.entry_archive.workflow)))
        return workflow_name

    def method_id_dft(self, settings_basis_set, functional_long_name: str):
        """Creates a method id for DFT calculations if all required data is
        present.
        """
        method_dict = RestrictedDict(
            mandatory_keys=[
                "program_name",
                "functional_long_name",
                "settings_basis_set",
                "scf_threshold_energy_change",
            ],
            optional_keys=(
                "smearing_kind",
                "smearing_width",
                "number_of_eigenvalues_kpoints",
            ),
            forbidden_values=[None]
        )
        method_dict['program_name'] = self.run.program.name

        # Functional settings
        method_dict['functional_long_name'] = functional_long_name

        # Basis set settings
        method_dict['settings_basis_set'] = settings_basis_set

        # k-point sampling settings if present. Add number of kpoints as
        # detected from eigenvalues. TODO: we would like to have info on the
        # _reducible_ k-point-mesh:
        #    - grid dimensions (e.g. [ 4, 4, 8 ])
        #    - or list of reducible k-points
        try:
            smearing_kind = self.repr_method.electronic.smearing.kind
            if smearing_kind is not None:
                method_dict['smearing_kind'] = smearing_kind
            smearing_width = self.repr_method.electronic.smearing.width
            if smearing_width is not None:
                smearing_width = '%.4f' % (smearing_width)
                method_dict['smearing_width'] = smearing_width
        except Exception:
            pass

        try:
            scc = self.run.calculation[-1]
            eigenvalues = scc.eigenvalues
            kpt = eigenvalues[-1].kpoints
        except (KeyError, IndexError):
            pass
        else:
            if kpt is not None:
                method_dict['number_of_eigenvalues_kpoints'] = str(len(kpt))

        # SCF convergence settings
        try:
            conv_thr = self.repr_method.scf.threshold_energy_change
            if conv_thr is not None:
                conv_thr = '%.13f' % (conv_thr.to(ureg.rydberg).magnitude)
                method_dict['scf_threshold_energy_change'] = conv_thr
        except Exception:
            pass

        # If all required information is present, safe the hash
        try:
            method_dict.check(recursive=True)
        except (KeyError, ValueError):
            pass
        else:
            return method_dict.hash()

    def equation_of_state_id(self, method_id: str, formula: str):
        """Creates an ID that can be used to group an equation of state
        calculation found within the same upload.
        """
        eos_dict = RestrictedDict(
            mandatory_keys=[
                "upload_id",
                "method_id",
                "formula",
            ],
            forbidden_values=[None]
        )

        # Only calculations from the same upload are grouped
        eos_dict['upload_id'] = self.entry_archive.metadata.upload_id

        # Method
        eos_dict["method_id"] = method_id

        # The formula should be same for EoS (maybe even symmetries)
        if self.material:
            eos_dict["formula"] = self.material.chemical_formula_hill

        # Form a hash from the dictionary
        try:
            eos_dict.check(recursive=True)
        except (KeyError, ValueError):
            pass
        else:
            return eos_dict.hash()

    def parameter_variation_id_dft(self, settings_basis_set, functional_long_name: str):
        """Creates an ID that can be used to group calculations that differ
        only by the used DFT parameters within the same upload.
        """
        # Create ordered dictionary with the values. Order is important for
        param_dict = RestrictedDict(
            mandatory_keys=[
                "upload_id",
                "program_name",
                "program_version",
                "settings_geometry",
                "functional_long_name",
                "scf_threshold_energy_change",
            ],
            optional_keys=(
                "atoms_pseudopotentials",
            ),
            forbidden_values=[None]
        )

        # Only calculations from the same upload are grouped
        param_dict['upload_id'] = self.entry_archive.metadata.upload_id

        # The same code and functional type is required
        param_dict['program_name'] = self.run.program.name
        param_dict['program_version'] = self.run.program.version

        # Get a string representation of the geometry. It is included as the
        # geometry should remain the same during parameter variation. By simply
        # using the atom labels and positions we assume that their
        # order/translation/rotation does not change.
        geom_dict: OrderedDict = OrderedDict()
        try:
            atoms = self.repr_system.atoms
            atom_labels = atoms.labels
            geom_dict['atom_labels'] = ', '.join(sorted(atom_labels))
            atom_positions = atoms['positions']
            geom_dict['atom_positions'] = np.array2string(
                atom_positions.to(ureg.angstrom).magnitude,  # convert to Angstrom
                formatter={'float_kind': lambda x: "%.6f" % x},  # type: ignore
            ).replace('\n', '')
            cell = atoms['lattice_vectors']
            geom_dict['simulation_cell'] = np.array2string(
                cell.to(ureg.angstrom).magnitude,  # convert to Angstrom
                formatter={'float_kind': lambda x: "%.6f" % x},  # type: ignore
            ).replace('\n', '')
        except Exception:
            pass
        param_dict['settings_geometry'] = geom_dict

        # TODO: Add other DFT-specific properties
        # considered variations:
        #   - smearing kind/width
        #   - k point grids
        #   - basis set parameters
        # convergence threshold should be kept constant during convtest
        param_dict['functional_long_name'] = functional_long_name
        conv_thr = self.repr_method.scf.threshold_energy_change if self.repr_method.scf is not None else None
        if conv_thr is not None:
            conv_thr = '%.13f' % (conv_thr.to(ureg.rydberg).magnitude)
        param_dict['scf_threshold_energy_change'] = conv_thr

        # Pseudopotentials are kept constant, if applicable
        if settings_basis_set is not None:
            pseudos = settings_basis_set.get('atoms_pseudopotentials', None)
            if pseudos is not None:
                param_dict['atoms_pseudopotentials'] = pseudos

        # Form a hash from the dictionary
        try:
            param_dict.check(recursive=True)
        except (KeyError, ValueError):
            pass
        else:
            return param_dict.hash()

    def basis_set_type(self) -> str:
        try:
            name = self.repr_method.basis_set[0].type
        except Exception:
            name = None
        if name:
            key = name.replace('_', '').replace('-', '').replace(' ', '').lower()
            name_mapping = {
                'gaussians': 'gaussians',
                'realspacegrid': 'real-space grid',
                'planewaves': 'plane waves'
            }
            name = name_mapping.get(key, name)
        return name

    def basis_set_name(self) -> Union[str, None]:
        try:
            name = self.repr_method.basis_set[0].name
        except Exception:
            name = None
        return name

    def core_electron_treatment(self) -> str:
        treatment = config.services.unavailable_value
        code_name = self.run.program.name
        if code_name is not None:
            core_electron_treatments = {
                'VASP': 'pseudopotential',
                'FHI-aims': 'full all electron',
                'exciting': 'full all electron',
                'quantum espresso': 'pseudopotential'
            }
            treatment = core_electron_treatments.get(code_name, config.services.unavailable_value)
        return treatment

    def xc_functional_names(self) -> Union[List[str], None]:
        if self.repr_method:
            functionals = set()
            try:
                for functional_type in ['exchange', 'correlation', 'hybrid', 'contributions']:
                    functionals.update([f.name for f in self.repr_method.dft.xc_functional[functional_type]])
            except Exception:
                pass
            if functionals:
                return sorted(functionals)
        return None

    def xc_functional_type(self, xc_functionals) -> str:
        if xc_functionals:
            name = xc_functionals[0]
            return xc_treatments.get(name[:3].lower(), config.services.unavailable_value)
        else:
            return config.services.unavailable_value

    def functional_long_name(self) -> str:
        """'Long' form of exchange-correlation functional, list of components
        and parameters as a string: see
        https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/XC-functional
        """
        xc_functional = MethodNormalizer.functional_long_name_from_method(self.repr_method, self.run.method)
        if xc_functional is config.services.unavailable_value:
            self.logger.warning(
                "Metainfo for 'XC_functional' not found, and could not "
                "compose name from 'section_XC_functionals'."
            )
        return xc_functional

    @staticmethod
    def functional_long_name_from_method(repr_method: Section, methods: List[Section]):
        """'Long' form of exchange-correlation functional, list of components
        and parameters as a string: see
        https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/XC-functional
        """
        if not repr_method or not methods:
            return None

        linked_methods = [repr_method]
        if repr_method.core_method_ref is not None:
            linked_methods.append(repr_method.core_method_ref)

        xc_functional = config.services.unavailable_value
        for method in linked_methods:
            if method.dft is None:
                continue
            try:
                sec_xc_functionals = method.dft.xc_functional
            except KeyError:
                pass
            else:
                if sec_xc_functionals is None:
                    return xc_functional
                components = {}
                for functional in ['exchange', 'correlation', 'hybrid', 'contributions']:
                    for component in sec_xc_functionals[functional]:
                        try:
                            cname = component.name
                        except KeyError:
                            pass
                        else:
                            this_component = ''
                            if component.weight is not None:
                                this_component = str(component.weight) + '*'
                            this_component += cname
                            components[cname] = this_component
                    result_array = []
                    for name in sorted(components):
                        result_array.append(components[name])
                    if len(result_array) >= 1:
                        xc_functional = '+'.join(result_array)
                if method.dft.xc_functional.name is None:
                    method.dft.xc_functional.name = xc_functional

        return xc_functional


def get_basis_set(entry_archive, repr_method, repr_system, logger) -> RestrictedDict:
    """Decide which type of basis set settings are applicable to the entry and
    return a corresponding settings as a RestrictedDict.

    Args:
        entry_archive: EntryArchive from which values are extracted.
        logger: Shared logger.

    Returns:
        RestrictedDict or None: Returns the extracted settings as a
        RestrictedDict. If no suitable basis set settings could be identified,
        returns None.
    """
    settings: BasisSet = None
    program_name = entry_archive.run[0].program.name
    if program_name == "exciting":
        settings = BasisSetExciting(entry_archive, repr_method, repr_system, logger)
    elif program_name == "FHI-aims":
        settings = BasisSetFHIAims(entry_archive, repr_method, repr_system, logger)
    else:
        return None

    return settings.to_dict()


class BasisSet(ABC):
    """Abstract base class for basis set settings. The idea is to create
    subclasses that inherit this class and hierarchically add new mandatory and
    optional settings with the setup()-function.
    """
    def __init__(self, entry_archive, repr_method, repr_system, logger):
        """
        """
        self._entry_archive = entry_archive
        self._repr_method = repr_method
        self._repr_system = repr_system
        self._logger = logger
        mandatory, optional = self.setup()
        self.settings = RestrictedDict(mandatory, optional, forbidden_values=[None])

    @abstractmethod
    def to_dict(self) -> RestrictedDict:
        """Used to extract basis set settings from the archive and returning
        them as a RestrictedDict.
        """
        pass

    @abstractmethod
    def setup(self) -> Tuple:
        """Used to define a list of mandatory and optional settings for a
        subclass.

        Returns:
            Should return a tuple of two lists: the first one defining
            mandatory keys and the second one defining optional keys.
        """
        mandatory: List = []
        optional: List = []
        return mandatory, optional


class BasisSetFHIAims(BasisSet):
    """Basis set settings for 'FHI-Aims' (code-dependent).
    """
    def setup(self) -> Tuple:
        # Get previously defined values from superclass
        mandatory, optional = super().setup()

        # Add new values
        mandatory += ["fhiaims_basis"]

        return mandatory, optional

    def to_dict(self):
        # Get basis set settings for each species
        aims_bs = self._repr_method.x_fhi_aims_section_controlIn_basis_set
        if not aims_bs:
            try:
                aims_bs = self._repr_method.method_ref.x_fhi_aims_section_controlIn_basis_set
            except Exception:
                pass
        if aims_bs is not None:
            bs_by_species = {}
            for this_aims_bs in aims_bs:
                this_bs_dict = self._values_to_dict(this_aims_bs, level=2)
                this_species = this_aims_bs['x_fhi_aims_controlIn_species_name'][0]
                bs_by_species[this_species] = this_bs_dict

            # Sort alphabetically by species label
            if bs_by_species:
                basis = OrderedDict()
                for k in sorted(bs_by_species.keys()):
                    basis[k] = bs_by_species[k]
                self.settings["fhiaims_basis"] = basis

        return self.settings

    @classmethod
    def _values_to_dict(cls, data, level=0):
        result = None
        if data is None:
            return None
        elif isinstance(data, (Section, dict)):
            result = OrderedDict()
            for k in sorted(cls._filtered_section_keys(data)):
                v = data.get(k, None)
                result[k] = cls._values_to_dict(v, level=level + 1)
        elif isinstance(data, (list)):
            result = []
            for k in range(len(data)):
                v = data[k]
                result.append(cls._values_to_dict(v, level=level + 1))
        elif isinstance(data, (np.ndarray)):
            result = data.tolist()
        else:
            result = data
        return result

    @classmethod
    def _filtered_section_keys(cls, section):
        for k in section.keys():
            # skip JSON-specific keys
            if k == '_gIndex':
                continue
            if k == '_name':
                continue
            else:
                # json values and subsections
                yield k


class BasisSetExciting(BasisSet):
    """Basis set settings for 'Exciting' (code-dependent).
    """
    def setup(self) -> Tuple:
        # Get previously defined values from superclass
        mandatory, optional = super().setup()

        # Add new values
        mandatory += [
            "muffin_tin_settings",
            "rgkmax",
            "gkmax",
            "lo",
            "lmaxapw",
        ]

        return mandatory, optional

    def to_dict(self):
        """Special case of basis set settings for Exciting code. See list at:
        https://gitlab.mpcdf.mpg.de/nomad-lab/encyclopedia-general/wikis/FHI-visit-preparation
        """
        # Add the muffin-tin settings for each species ordered alphabetically by atom label
        try:
            groups = self._repr_system.x_exciting_section_atoms_group
            groups = sorted(groups, key=lambda group: group.x_exciting_geometry_atom_labels)
            muffin_tin_settings = OrderedDict()
            for group in groups:
                label = group.x_exciting_geometry_atom_labels
                try:
                    muffin_tin_settings["{}_muffin_tin_radius".format(label)] = "%.6f" % (group.x_exciting_muffin_tin_radius.to(ureg.angstrom).magnitude)
                except Exception:
                    muffin_tin_settings["{}_muffin_tin_radius".format(label)] = None
                try:
                    muffin_tin_settings["{}_muffin_tin_points".format(label)] = "%d" % group.x_exciting_muffin_tin_points
                except Exception:
                    muffin_tin_settings["{}_muffin_tin_points".format(label)] = None
            self.settings["muffin_tin_settings"] = muffin_tin_settings
        except Exception:
            pass

        # Other important method settings
        system = self._repr_system
        try:
            self.settings['rgkmax'] = "%.6f" % (system.x_exciting_rgkmax.magnitude)
        except Exception:
            pass
        try:
            self.settings['gkmax'] = "%.6f" % (1e-10 * system.x_exciting_gkmax.magnitude)
        except Exception:
            pass
        try:
            self.settings['lo'] = "%d" % (system.x_exciting_lo)
        except Exception:
            pass
        try:
            self.settings['lmaxapw'] = "%d" % (system.x_exciting_lmaxapw)
        except Exception:
            pass

        return self.settings



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from ase import Atoms
from nomad.datamodel.metainfo.simulation.calculation import Calculation

from nomad.normalizing.normalizer import Normalizer
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.workflow import (
    Workflow, Task, SinglePoint, GeometryOptimization, MolecularDynamics, Phonon, Elastic,
    Thermodynamics)


def resolve_difference(values):
    delta_values = None

    values = [v for v in values if v is not None]
    for n in range(-1, -len(values), -1):
        delta_values = abs(values[n] - values[n - 1])
        if delta_values != 0.0:
            break

    return delta_values


class TaskNormalizer(Normalizer):
    def __init__(self, entry_archive, workflow_index):
        super().__init__(entry_archive)
        workflow_index = workflow_index if len(entry_archive.workflow) < workflow_index else -1
        self.workflow = entry_archive.workflow[workflow_index]
        run = self.workflow.run_ref
        self.run = run if run else entry_archive.run[-1]
        self._serial_tasks = ['geometry_optimization', 'molecular_dynamics']

    def normalize(self):
        if self.workflow.task:
            return

        # resolve references to calculations/workflows
        # TODO serialization of external archives does not work.
        references = []
        if self.workflow.workflows_ref:
            references = self.workflow.workflows_ref
        elif self.workflow.calculations_ref and not references:
            references = self.workflow.calculations_ref
        if self.workflow.type in self._serial_tasks:
            # for serial tasks such as geometry_optimization (GO), calculations (C) /
            # single point (SP) workflows are run in series, i.e. input of a step is the
            # preceeding step. It starts from the input structure in GO and ends from the
            # relaxed calculation reported to the the GO workflow. This is shown in the
            # following schematic for a GO workflow:
            #
            # GO <----------------------------
            #  |                             |
            #  v                             |
            #  SP1/C1 -> SP2/C2 -> .... -> SPN/CN
            references += [self.workflow]
            for n, reference in enumerate(references):
                sec_task = self.workflow.m_create(Task)
                input = references[-1] if n == 0 else references[n - 1]
                sec_task.input_workflow = input if isinstance(input, Workflow) else None
                sec_task.input_calculation = input if isinstance(input, Calculation) else None
                sec_task.output_workflow = reference if isinstance(reference, Workflow) else None
                sec_task.output_calculation = reference if isinstance(reference, Calculation) else None
        else:
            # for parallel tasks such as phonon workflow (Ph), calculations (C)/ single
            # point workflows (SP) are run in parallel and for each run, two tasks are
            # realized, first is when the structure is generated by the workflow engine
            # (input) and run by the energy/force calculator (output) and the second is
            # when the results (input) are fed to the workflow engine (output). We then
            # have twice as number of tasks as workflows/calculations. The schematic of a
            # phonon workflow, for example is shown as follows:
            #
            # Ph -------------...
            #  |^       |^       |^
            #  v|       v|       v|
            #  SP1/C1   SP2/C2   SPN/CN
            for reference in references:
                for n in range(2):
                    input, output = (self.workflow, reference) if n == 0 else (reference, self.workflow)
                    sec_task = self.workflow.m_create(Task)
                    sec_task.input_workflow = input if isinstance(input, Workflow) else None
                    sec_task.input_calculation = input if isinstance(input, Calculation) else None
                    sec_task.output_workflow = output if isinstance(output, Workflow) else None
                    sec_task.output_calculation = output if isinstance(output, Calculation) else None


class SinglePointNormalizer(TaskNormalizer):
    def normalize(self):
        super().normalize()
        self.section = self.workflow.single_point
        if not self.section:
            self.section = self.workflow.m_create(SinglePoint)

        if not self.section.method:
            try:
                method = self.run.method[-1]
                self.section.method = method.electronic.method
            except Exception:
                pass

        scc = self.run.calculation
        if not scc:
            return

        if not self.section.n_scf_steps:
            self.section.n_scf_steps = len(scc[-1].scf_iteration)

        energies = [scf.energy.total.value for scf in scc[-1].scf_iteration if scf.energy is not None and scf.energy.total is not None]
        delta_energy = resolve_difference(energies)
        if not self.section.final_scf_energy_difference and delta_energy is not None:
            self.section.final_scf_energy_difference = delta_energy

        if not self.section.is_converged and delta_energy is not None:
            try:
                threshold = self.run.method[-1].scf.threshold_energy_change
                self.section.is_converged = bool(delta_energy <= threshold)
            except Exception:
                pass

        if not self.section.with_density_of_states:
            self.section.with_density_of_states = len(scc[-1].dos_electronic) > 0

        if not self.section.with_bandstructure:
            self.section.with_bandstructure = len(scc[-1].band_structure_electronic) > 0

        if not self.section.with_eigenvalues:
            self.section.with_eigenvalues = len(scc[-1].eigenvalues) > 0

        if not self.section.with_volumetric_data:
            self.section.with_volumetric_data = (
                len(scc[-1].potential) > 0 or len(scc[-1].density_charge) > 0)

        if not self.section.with_excited_states:
            self.section.with_excited_states = len(scc[-1].excited_states) > 0


class GeometryOptimizationNormalizer(TaskNormalizer):

    def _get_geometry_optimization_type(self):
        sec_system = self.run.system
        if not sec_system:
            return

        def compare_cell(cell1, cell2):
            if (cell1 == cell2).all():
                return None
            else:
                cell1_normed = cell1 / np.linalg.norm(cell1)
                cell2_normed = cell2 / np.linalg.norm(cell2)
                if (cell1_normed == cell2_normed).all():
                    return 'cell_volume'
                else:
                    return 'cell_shape'

        if len(sec_system) < 2:
            return 'static'

        else:
            if sec_system[0].atoms is None or sec_system[-1].atoms is None:
                return 'static'

            cell_init = sec_system[0].atoms.lattice_vectors
            cell_final = sec_system[-1].atoms.lattice_vectors
            if cell_init is None or cell_final is None:
                return 'static'

            cell_relaxation = compare_cell(cell_init.magnitude, cell_final.magnitude)

            if cell_relaxation is not None:
                return cell_relaxation

            atom_pos_init = sec_system[0].atoms.positions
            atom_pos_final = sec_system[-1].atoms.positions

            if atom_pos_init is None or atom_pos_final is None:
                return 'static'

            if (atom_pos_init.magnitude == atom_pos_final.magnitude).all():
                return 'static'

            return 'ionic'

    def normalize(self):
        super().normalize()
        self.section = self.workflow.geometry_optimization
        if self.section is None:
            self.section = self.workflow.m_create(GeometryOptimization)

        if not self.section.type:
            try:
                geometry_optimization_type = self._get_geometry_optimization_type()
                self.section.type = geometry_optimization_type
            except Exception:
                pass

        if not self.section.optimization_steps:
            scc = self.run.calculation
            self.section.optimization_steps = len(scc)

        if not self.section.final_energy_difference:
            energies = []
            for scc in self.run.calculation:
                if scc.energy is not None and scc.energy.total is not None:
                    energies.append(scc.energy.total.value)

            delta_energy = resolve_difference(energies)
            if delta_energy is not None:
                self.section.final_energy_difference = delta_energy

        if not self.section.final_force_maximum:
            scc = self.run.calculation
            if len(scc) > 0:
                if scc[-1].forces is not None and scc[-1].forces.total is not None:
                    forces = scc[-1].forces.total.value
                    if forces is not None:
                        max_force = np.max(np.linalg.norm(forces.magnitude, axis=1))
                        self.section.final_force_maximum = max_force * forces.units

        # Store the energies as an explicit list. If a step within the
        # trajectory does not contain an energy the rest of the energies in the
        # trajectory are not included.
        trajectory = self.workflow.calculations_ref
        if trajectory:
            n_steps = len(trajectory)
            energies = []
            invalid = False
            for step in range(n_steps):
                try:
                    energy = trajectory[step].energy.total.value
                except (IndexError, AttributeError):
                    invalid = True
                    break
                energies.append(energy.magnitude)
            if invalid:
                self.logger.warning("energy not reported for an scc that is part of a geometry optimization")
            if energies:
                self.section.energies = energies

        if not self.section.final_displacement_maximum:
            try:
                system = self.run.system
                ase_atoms = [Atoms(
                    positions=system_i.atoms.positions.magnitude,
                    cell=system_i.atoms.lattice_vectors.magnitude,
                    pbc=system_i.atoms.periodic) for system_i in system]
                _ = [atoms.wrap() for atoms in ase_atoms]
                displacements = [np.max(np.abs(
                    ase_atoms[n].get_positions() - ase_atoms[n - 1].get_positions())) for n in range(1, len(ase_atoms))]
                self.section.final_displacement_maximum = displacements[np.amax(np.nonzero(displacements))]
            except Exception:
                pass

        if not self.section.is_converged_geometry:
            # we can have several criteria for convergence: energy, force, displacement
            criteria = []
            try:
                criteria.append(self.section.final_energy_difference <= self.section.convergence_tolerance_energy_difference)
            except Exception:
                pass

            try:
                criteria.append(self.section.final_force_maximum <= self.section.convergence_tolerance_force_maximum)
            except Exception:
                pass

            try:
                criteria.append(self.section.final_displacement_maximum <= self.section.convergence_tolerance_displacement_maximum)
            except Exception:
                pass

            # converged when either criterion is met
            if criteria:
                self.section.is_converged_geometry = True in criteria


class PhononNormalizer(TaskNormalizer):
    def _get_n_imaginary_frequencies(self):
        scc = self.run.calculation
        if not scc:
            return
        sec_band = scc[0].band_structure_phonon
        if not sec_band:
            return
        result = 0
        for band_segment in sec_band[0].segment:
            freq = band_segment.energies.magnitude
            result += np.count_nonzero(np.array(freq) < 0)
        return result

    def normalize(self):
        super().normalize()
        self.section = self.workflow.phonon

        if not self.section:
            self.section = self.workflow.m_create(Phonon)

        if not self.section.n_imaginary_frequencies:
            # get number from bands (not complete as this is not the whole mesh)
            self.section.n_imaginary_frequencies = self._get_n_imaginary_frequencies()


class ElasticNormalizer(TaskNormalizer):
    def _resolve_mechanical_stability(self):
        spacegroup, c = None, None
        try:
            spacegroup = self.run.system[-1].symmetry[-1].space_group_number
            c = self.section.elastic_constants_matrix_second_order
        except Exception:
            return False

        if c is None or spacegroup is None:
            return False

        # see Phys. Rev B 90, 224104 (2014)
        res = False
        if spacegroup <= 2:  # Triclinic
            res = np.count_nonzero(c < 0)
        elif spacegroup <= 15:  # Monoclinic
            res = np.count_nonzero(c < 0)
        elif spacegroup <= 74:  # Orthorhombic
            res =\
                c[0][0] > 0 and c[0][0] * c[1][1] > c[0][1] ** 2 and\
                c[0][0] * c[1][1] * c[2][2] + 2 * c[0][1] * c[0][2] * c[1][2] -\
                c[0][0] * c[1][2] ** 2 - c[1][1] * c[0][2] ** 2 - c[2][2] * c[0][1] ** 2 > 0 and\
                c[3][3] > 0 and c[4][4] > 0 and c[5][5] > 0
        elif spacegroup <= 88:  # Tetragonal II
            res =\
                c[0][0] > abs(c[0][1]) and\
                2 * c[0][2] ** 2 < c[2][2] * (c[0][0] + c[0][1])
        elif spacegroup <= 142:  # Tetragonal I
            res =\
                c[0][0] > abs(c[0][1]) and\
                2 * c[0][2] ** 2 < c[2][2] * (c[0][0] + c[0][1]) and\
                c[3][3] > 0 and c[5][5] > 0
        elif spacegroup <= 148:  # rhombohedral II
            res =\
                c[0][0] > abs(c[0][1]) and c[3][3] > 0 and\
                c[0][2] ** 2 < (0.5 * c[2][2] * (c[0][0] + c[0][1])) and\
                c[0][3] ** 2 + c[0][4] ** 2 < 0.5 * c[3][3] * (c[0][0] - c[0][1])
        elif spacegroup <= 167:  # rhombohedral I
            res =\
                c[0][0] > abs(c[0][1]) and c[3][3] > 0 and\
                c[0][2] ** 2 < 0.5 * c[2][2] * (c[0][0] + c[0][1]) and\
                c[0][3] ** 2 < 0.5 * c[3][3] * (c[0][0] - c[0][1])
        elif spacegroup <= 194:  # hexagonal I
            res =\
                c[0][0] > abs(c[0][1]) and\
                2 * c[0][2] ** 2 < c[2][2] * (c[0][0] + c[0][1]) and\
                c[3][3] > 0 and c[5][5] > 0
        else:  # cubic
            res = c[0][0] - c[0][1] > 0 and c[0][0] + 2 * c[0][1] > 0 and c[3][3] > 0

        return res

    def _get_maximum_fit_error(self):
        max_error = 0.0
        if len(self.run.calculation) == 0:
            return max_error

        for diagram in self.section.strain_diagrams:
            if diagram.type == 'cross-validation':
                error = np.amax(diagram.value)
                max_error = error if error > max_error else max_error

        return max_error

    def normalize(self):
        super().normalize()
        self.section = self.workflow.elastic

        if not self.section:
            self.section = self.workflow.m_create(Elastic)

        if self.section.is_mechanically_stable is None:
            self.section.is_mechanically_stable = bool(self._resolve_mechanical_stability())

        if self.section.fitting_error_maximum is None:
            self.section.fitting_error_maximum = self._get_maximum_fit_error()


class MolecularDynamicsNormalizer(TaskNormalizer):
    def _is_with_thermodynamics(self):
        try:
            return len(self.run.calculation[-1].thermodynamics[-1].values()) > 0
        except Exception:
            return False

    def _is_with_trajectory(self):
        try:
            return self.run.system[-1].atoms.positions is not None
        except Exception:
            return False

    def normalize(self):
        super().normalize()
        self.section = self.workflow.molecular_dynamics

        if not self.section:
            self.section = self.workflow.m_create(MolecularDynamics)

        if self.section.with_thermodynamics is None:
            self.section.with_thermodynamics = self._is_with_thermodynamics()

        if self.section.with_trajectory is None:
            self.section.with_trajectory = self._is_with_trajectory()


class ThermodynamicsNormalizer(TaskNormalizer):
    def normalize(self):
        super().normalize()
        self.section = self.workflow.thermodynamics

        if not self.run.calculation or not self.run.calculation[0].thermodynamics:
            return

        if not self.section:
            self.section = self.workflow.m_create(Thermodynamics)

        def set_thermo_property(name):
            values = []
            quantity = None
            for scc in self.run.calculation:
                try:
                    for thermo in scc.thermodynamics:
                        quantity = thermo[name]
                        values.append(quantity.magnitude if hasattr(quantity, 'magnitude') else quantity)
                except Exception:
                    pass
            if len(values) == 0:
                return

            unit = quantity.units if hasattr(quantity, 'units') else 1.0
            setattr(self.section, name, np.array(values) * unit)

        if self.section.temperature is None:
            set_thermo_property('temperature')

        if self.section.pressure is None:
            set_thermo_property('pressure')

        if self.section.helmholtz_free_energy is None:
            set_thermo_property('helmholtz_free_energy')

        if self.section.vibrational_free_energy_at_constant_volume is None:
            set_thermo_property('vibrational_free_energy_at_constant_volume')

        if self.section.heat_capacity_c_v is None:
            set_thermo_property('heat_capacity_c_v')

        # TODO add values for specific energy


class WorkflowNormalizer(Normalizer):
    '''
    This normalizer produces information specific to a workflow.
    '''
    def __init__(self, entry_archive):
        super().__init__(entry_archive)
        self._elastic_programs = ['elastic']
        self._phonon_programs = ['phonopy']
        self._molecular_dynamics_programs = ['lammps']

    def _resolve_workflow_type(self, run):
        # resolve it from parser
        workflow_type = None
        try:
            program_name = run.program.name
        except Exception:
            return

        if program_name:
            program_name = program_name.lower()

        if program_name in self._elastic_programs:
            workflow_type = 'elastic'

        elif program_name in self._molecular_dynamics_programs:
            workflow_type = 'molecular_dynamics'

        elif program_name in self._phonon_programs:
            workflow_type = 'phonon'

        # resolve if from scc
        if workflow_type is None:
            if len(run.calculation) == 1:
                workflow_type = 'single_point'
            else:
                workflow_type = 'geometry_optimization'

        return workflow_type

    def normalize(self, logger=None) -> None:
        super().normalize()

        # Do nothing if section_run is not present
        if not self.entry_archive.run:
            return

        if not self.entry_archive.workflow:
            self.entry_archive.m_create(Workflow)

        for n, sec_workflow in enumerate(self.entry_archive.workflow):
            # we get reference the referenced run from which information can be extracted
            sec_run = sec_workflow.run_ref
            sec_run = sec_run if sec_run else self.entry_archive.run[-1]

            scc = sec_run.calculation
            if not sec_workflow.calculation_result_ref:
                if scc:
                    sec_workflow.calculation_result_ref = scc[-1]

            if not sec_workflow.calculations_ref:
                if scc:
                    sec_workflow.calculations_ref = scc

            if sec_workflow.type is None:
                workflow_type = self._resolve_workflow_type(sec_run)
                sec_workflow.type = workflow_type

            if sec_workflow.type == 'geometry_optimization':
                GeometryOptimizationNormalizer(self.entry_archive, n).normalize()

            elif sec_workflow.type == 'phonon':
                PhononNormalizer(self.entry_archive, n).normalize()

            elif sec_workflow.type == 'elastic':
                ElasticNormalizer(self.entry_archive, n).normalize()

            elif sec_workflow.type == 'molecular_dynamics':
                MolecularDynamicsNormalizer(self.entry_archive, n).normalize()

            elif sec_workflow.type == 'single_point':
                SinglePointNormalizer(self.entry_archive, n).normalize()

            # add thermodynamics data
            ThermodynamicsNormalizer(self.entry_archive, n).normalize()

            # remove the section workflow again, if the parser/normalizer could not produce a result
            if sec_workflow.calculation_result_ref is None:
                self.entry_archive.m_remove_sub_section(EntryArchive.workflow, n)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import logging
import numpy as np
from datetime import datetime

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.simulation.run import Run, Program, TimeRun
from nomad.datamodel.metainfo.simulation.system import (
    System, Atoms
)
from nomad.datamodel.metainfo.simulation.method import (
    Method, DFT, XCFunctional, Functional
)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, ScfIteration, Forces, ForcesEntry, Stress,
    StressEntry, Multipoles, MultipolesEntry
)
from .metainfo.qbox import x_qbox_section_MLWF


re_n = r'[\n\r]'
re_f = r'[-+]?\d+\.\d*(?:[DdEe][-+]\d+)?'


class QboxXMLParser(TextParser):
    def init_quantities(self):
        run_quantities = [
            Quantity(
                'parameter',
                r'\[qbox\] \<cmd\>set +(.+?)\<\/cmd\>',
                str_operation=lambda x: x.split(' ', 1), repeats=True),
            Quantity(
                'species',
                r'(\<species name=[\s\S]=?)</species>',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity('symbol', r'\<symbol\>([A-Z][a-z]*)', dtype=str),
                    Quantity('atomic_numer', r'\<atomic_number\>(\d+)', dtype=np.int32),
                    Quantity('mass', r'\<mass\>(\d+)', dtype=np.float64)
                ])
            ),
            Quantity(
                'iteration',
                r'(\<iteration[\s\S]+?)</iteration>',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity('charge', rf'total_electronic_charge: +({re_f})', dtype=np.float64),
                    Quantity('energy_kinetic_electronic', rf'\<ekin\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_conf', rf'\<econf\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_ps', rf'\<eps\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_nl', rf'\<enl\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_coulomb', rf'\<ecoul\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_xc', rf'\<exc\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_sr', rf'\<esr\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_self', rf'\<eself\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_ts', rf'\<ets\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_exf', rf'\<eexf\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_total', rf'\<etotal\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_pv', rf'\<epv\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_efield', rf'\<eefield\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity('energy_x_qbox_enthalpy', rf'\<enthalpy\> +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity(
                        'stress_tensor',
                        r'(\<stress_tensor[\s\S]+?)\<\/stress_tensor\>',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'contribution',
                                r'(\<sigma_.*\_*xx\>[\s\S]+?)\<\/sigma_.*\_*xz\>',
                                repeats=True, sub_parser=TextParser(quantities=[
                                    Quantity('kind', r'\<sigma_e(.+?)\_', dtype=str),
                                    Quantity('value', rf'\> +({re_f})', dtype=np.float64, repeats=True)
                                ])
                            )
                        ])
                    ),
                    Quantity(
                        'multipole',
                        r'(\<[a-z]+pole\>[\s\S]+?)\<\/[a-z]+pole\>',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity(
                                'dipole',
                                rf'\<dipole_([a-z]+)\> +({re_f} +{re_f} +{re_f})',
                                repeats=True
                            ),
                            Quantity(
                                'quadrupole',
                                rf'\<quadrupole_([a-z]+)\>\s+'
                                rf'({re_f} +{re_f} +{re_f})\s+'
                                rf'({re_f} +{re_f} +{re_f})\s+'
                                rf'({re_f} +{re_f} +{re_f})\s+',
                                repeats=True
                            )
                        ])
                    ),
                    Quantity(
                        'mlwf',
                        r'(\<mlwf_set[\s\S]+?)\<\/mlwf_set\>',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'center',
                                rf'mlwf center=\" +({re_f} +{re_f} +{re_f}) +\"\s+spread=\" +({re_f})',
                                repeats=True, dtype=np.dtype(np.float64)
                            )
                        ])
                    ),
                    Quantity(
                        'scf',
                        r'(total_electronic_charge:.+\s+\<eigenvalue_sum[\s\S]+?)\<\/etotal_int\>',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity('charge', rf'total_electronic_charge: +({re_f})', dtype=np.float64),
                            Quantity(
                                'energy_sum_eigenvalues',
                                rf'\<eigenvalue_sum\> +({re_f})',
                                unit=ureg.hartree, dtype=np.float64),
                            Quantity(
                                'energy_total',
                                rf'\<etotal_int\> +({re_f})',
                                unit=ureg.hartree, dtype=np.float64),
                        ])
                    ),
                    Quantity(
                        'atomset',
                        r'(\<atomset\>[\s\S]+?)</atomset>',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'lattice_vectors',
                                rf'\<unit_cell\s+'
                                rf'a=\" +({re_f}) +({re_f}) +({re_f})\"\s+'
                                rf'b=\" +({re_f}) +({re_f}) +({re_f})\"\s+'
                                rf'c=\" +({re_f}) +({re_f}) +({re_f})\"\s+',
                                dtype=np.dtype(np.float64), shape=(3, 3), unit=ureg.bohr
                            ),
                            Quantity(
                                'atom',
                                r'(\<atom[\s\S]+?)\<\/atom\>',
                                repeats=True, sub_parser=TextParser(quantities=[
                                    Quantity(
                                        'label',
                                        r'\<atom name=\"([A-Z][a-z]*)',
                                        dtype=str
                                    ),
                                    Quantity(
                                        'position',
                                        rf'\<position\> +({re_f} +{re_f} +{re_f})',
                                        dtype=np.dtype(np.float64)
                                    ),
                                    Quantity(
                                        'velocity',
                                        rf'\<velocity\> +({re_f} +{re_f} +{re_f})',
                                        dtype=np.dtype(np.float64)
                                    ),
                                    Quantity(
                                        'force',
                                        rf'\<force\> +({re_f} +{re_f} +{re_f})',
                                        dtype=np.dtype(np.float64)
                                    )
                                ])
                            ),

                        ]))
                ])
            )
        ]

        self._quantities = [
            Quantity('program_version', r'I qbox (\S+)', dtype=str),
            Quantity('start_time', r'\<start_time\> (\d\d\d\d\-\d\d\-\d\dT\d\d:\d\d:\d\dZ)', dtype=str),
            Quantity('end_time', r'\<end_time\> (\d\d\d\d\-\d\d\-\d\dT\d\d:\d\d:\d\dZ)', dtype=str),
            Quantity('x_qbox_nodename', r'\<nodename\> (.+) \<\/nodename\>', dtype=str, flatten=False),
            Quantity('x_qbox_loading_xml_file', r'LoadCmd: loading from (\S+)', dtype=str),
            Quantity(
                'run',
                r'(\[qbox\] \<cmd\>[\s\S]+?\<timing name=\" +charge_vft.+)',
                repeats=True, sub_parser=TextParser(quantities=run_quantities)
            )
        ]


class QboxParser:
    def __init__(self):
        self.out_parser = QboxXMLParser()
        self._xc_map = {
            'LDA': ['LDA_X', 'LDA_C_PZ'],
            'VMN': ['LDA_X', 'LDA_C_VWN'],
            'PBE': ['GGA_X_PBE', 'GGA_C_PBE'],
            'PBE0': ['GGA_X_PBE', 'GGA_C_PBE'],
            'B3LYP': ['HYB_GGA_XC_B3LYP5']
        }

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.out_parser.logger = self.logger

    def parse_run(self, index):
        sec_run = self.archive.m_create(Run)
        sec_run.program = Program(name='qbox', version=self.out_parser.get('program_version'))

        def format_time(time):
            if time is not None:
                return datetime.strptime(time, '%Y-%m-%dT%H:%M:%S%z').timestamp()

        sec_run.time_run = TimeRun(
            date_start=format_time(self.out_parser.start_time),
            date_end=format_time(self.out_parser.end_time))

        sec_method = sec_run.m_create(Method)
        parameters = {key: val for key, val in self.out_parser.run[index].get('parameter', [])}
        sec_method.x_qbox_input_parameters = parameters
        if parameters.get('xc') is not None:
            sec_method.dft = DFT(xc_functional=XCFunctional())
            for xc_functional in self._xc_map.get(parameters['xc'], []):
                if '_X_' in xc_functional or xc_functional.endswith('_X'):
                    sec_method.dft.xc_functional.exchange.append(Functional(name=xc_functional))
                elif '_C_' in xc_functional or xc_functional.endswith('_C'):
                    sec_method.dft.xc_functional.correlation.append(Functional(name=xc_functional))
                elif 'HYB' in xc_functional:
                    sec_method.dft.xc_functional.hybrid.append(Functional(name=xc_functional))
                else:
                    sec_method.dft.xc_functional.contributions.append(Functional(name=xc_functional))

        def symmetrize_stress(stress):
            if not stress:
                return
            symmetrized = np.zeros((3, 3))
            symmetrized[0][0] = stress[0]
            symmetrized[1][1] = stress[1]
            symmetrized[2][2] = stress[2]
            symmetrized[0][1] = symmetrized[1][0] = stress[3]
            symmetrized[1][2] = symmetrized[2][1] = stress[4]
            symmetrized[0][2] = symmetrized[2][0] = stress[5]
            return symmetrized * ureg.GPa

        for iteration in self.out_parser.run[index].get('iteration', []):
            sec_scc = sec_run.m_create(Calculation)
            sec_scc.energy = Energy()
            sec_system = sec_run.m_create(System)
            sec_scc.system_ref = sec_system

            for key, val in iteration.items():
                if key.startswith('energy_') and val is not None:
                    setattr(sec_scc.energy, key.replace('energy_', ''), EnergyEntry(value=val))

            if iteration.stress_tensor is not None:
                sec_stress = sec_scc.m_create(Stress)
                for contribution in iteration.stress_tensor.get('contribution', []):
                    value = symmetrize_stress(contribution.get('value', []))
                    if contribution.kind is None or contribution.kind == 'ks':
                        sec_stress.total = StressEntry(value=value)
                    else:
                        sec_stress.contributions.append(StressEntry(
                            kind=contribution.kind,
                            value=value
                        ))

            if iteration.multipole is not None:
                sec_multipoles = sec_scc.m_create(Multipoles)
                for multipole in iteration.multipole:
                    # TODO add contributions
                    for dipole in multipole.get('dipole', []):
                        if dipole[0] == 'total':
                            sec_multipoles.dipole = MultipolesEntry(total=np.array(dipole[1:], dtype=np.float64))
                    for quadrupole in multipole.get('quadrupole', []):
                        if quadrupole[0] == 'total':
                            sec_multipoles.quadrupole = MultipolesEntry(total=np.array(quadrupole[1:], dtype=np.float64))

            if iteration.mlwf is not None:
                for center in iteration.mlwf.get('center', []):
                    sec_mlwf = sec_scc.m_create(x_qbox_section_MLWF)
                    sec_mlwf.x_qbox_geometry_MLWF_atom_positions = center[:3] * ureg.bohr
                    sec_mlwf.x_qbox_geometry_MLWF_atom_spread = center[3] * ureg.bohr

            for scf in iteration.get('scf', []):
                sec_scf = sec_scc.m_create(ScfIteration)
                sec_scf.energy = Energy(
                    total=EnergyEntry(value=scf.energy_total),
                    sum_eigenvalues=EnergyEntry(value=scf.energy_sum_eigenvalues))

            if iteration.atomset is not None:
                atoms = iteration.atomset.get('atom', [])
                sec_system.atoms = Atoms(
                    lattice_vectors=iteration.atomset.lattice_vectors,
                    positions=[atom.position for atom in atoms] * ureg.bohr,
                    labels=[atom.label for atom in atoms],
                    velocities=[atom.velocity for atom in atoms] * (ureg.bohr / ureg.atomic_unit_of_time))
                forces = [atom.force for atom in atoms]
                if forces:
                    sec_scc.forces = Forces(total=ForcesEntry(value=forces * (ureg.hartree / ureg.bohr)))

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger
        self.maindir = os.path.dirname(self.filepath)
        self.init_parser()

        for n in range(len(self.out_parser.get('run', []))):
            self.parse_run(n)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import io
import typing
import sys

from nomad import config, utils, datamodel

from .api import Auth


def parse(
        mainfile_path: str,
        parser_name: str = None,
        backend_factory: typing.Callable = None,
        strict: bool = True, logger=None) -> typing.List[datamodel.EntryArchive]:
    '''
    Run the given parser on the provided mainfile. If parser_name is given, we only try
    to match this parser, otherwise we try to match all parsers.
    '''
    from nomad import parsing
    from nomad.parsing import parsers
    mainfile = os.path.basename(mainfile_path)

    if logger is None:
        logger = utils.get_logger(__name__)

    mainfile_path = os.path.abspath(mainfile_path)
    parser, mainfile_keys = parsers.match_parser(mainfile_path, strict=strict, parser_name=parser_name)
    if isinstance(parser, parsing.MatchingParser):
        parser_name = parser.name
    else:
        parser_name = parser.__class__.__name__

    assert parser is not None, 'there is no parser matching %s' % mainfile
    logger = logger.bind(parser=parser.name)  # type: ignore
    logger.info('identified parser')
    if hasattr(parser, 'backend_factory'):
        setattr(parser, 'backend_factory', backend_factory)

    entry_archives = parsers.run_parser(mainfile_path, parser, mainfile_keys, logger)

    logger.info('ran parser')
    return entry_archives


def normalize(
        normalizer: typing.Union[str, typing.Callable], entry_archive, logger=None):
    from nomad import normalizing

    if logger is None:
        logger = utils.get_logger(__name__)

    if isinstance(normalizer, str):
        normalizer = next(
            normalizer_instance for normalizer_instance in normalizing.normalizers
            if normalizer_instance.__class__.__name__ == normalizer)

    assert normalizer is not None, 'there is no normalizer %s' % str(normalizer)
    normalizer_instance = typing.cast(typing.Callable, normalizer)(entry_archive)
    logger = logger.bind(normalizer=normalizer_instance.__class__.__name__)
    logger.info('identified normalizer')

    normalizer_instance.normalize(logger=logger)
    logger.info('ran normalizer')


def normalize_all(entry_archive, logger=None):
    '''
    Parse the downloaded entry and run the whole normalizer chain.
    '''
    from nomad import normalizing

    for normalizer in normalizing.normalizers:
        if normalizer.domain is None or normalizer.domain == entry_archive.metadata.domain:
            normalize(normalizer, entry_archive, logger=logger)


class LocalEntryProcessing:
    '''
    Instances represent a local reproduction of the processing for a single entry.
    It allows to download raw data from a nomad server and reproduce its processing
    (parsing, normalizing) with the locally installed parsers and normalizers.

    The use-case is error/warning reproduction. Use ELK to identify errors, use
    the upload, archive ids to given by ELK, and reproduce and fix the error
    in your development environment.

    Arguments:
        entry_id: The entry_id of the entry to locally process.
        override: Set to true to override any existing local entry data.
        auth: Optional Auth object to download private data.
    '''
    def __init__(self, entry_id: str, override: bool = False, auth: Auth = None) -> None:
        from nomad import files
        from nomad.client import api

        self.entry_id = entry_id
        response = self.__handle_response(
            api.get(f'entries/{self.entry_id}', auth=auth))
        self.mainfile = response.json()['data']['mainfile']
        self.upload_id = response.json()['data']['upload_id']

        self.local_path = os.path.join(config.fs.tmp, f'repro_{self.entry_id}.zip')
        if not os.path.exists(os.path.dirname(self.local_path)):
            os.makedirs(os.path.dirname(self.local_path))

        if not os.path.exists(self.local_path) or override:
            # download raw if not already downloaded or if override is set
            print('Downloading', self.entry_id)
            response = self.__handle_response(
                api.get(f'entries/{self.entry_id}/raw', auth=auth))

            with open(self.local_path, 'wb') as f:
                for chunk in response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE):
                    f.write(chunk)
        else:
            print('Entry already downloaded.')

        self.upload_files = files.StagingUploadFiles(upload_id=f'tmp_{self.entry_id}', create=True)

    def __handle_response(self, response):
        if response.status_code == 404:
            print('Could not find entry, maybe it was deleted or the ids are wrong:', file=sys.stderr)
            print(response.text, file=sys.stderr)
            sys.exit(1)
        elif response.status_code != 200:
            print(f'API error: {response.text}', file=sys.stderr)
            sys.exit(1)

        return response

    def __enter__(self):
        # open/extract upload file
        print('Extracting entry data.')
        if self.upload_files.is_empty():  # Only add the files once
            self.upload_files.add_rawfiles(self.local_path)

        for raw_file in self.upload_files.raw_directory_list(recursive=True):
            if raw_file.path == f'{self.upload_id}/{self.mainfile}':
                print('Identified mainfile.')
                return self

        print('Count not find file in download.', file=sys.stderr)
        sys.exit(1)

    def __exit__(self, exception_type, exception, trace):
        if exception:
            import traceback
            traceback.print_exc()
        try:
            self.upload_files.delete()
        except Exception:
            pass
        if exception:
            sys.exit(1)

    def parse(self, parser_name: str = None, **kwargs) -> typing.List[datamodel.EntryArchive]:
        '''
        Run the given parser on the downloaded entry. If no parser is given,
        do parser matching and use the respective parser.
        '''
        return parse(
            self.upload_files.raw_file_object(f'{self.upload_id}/{self.mainfile}').os_path,
            parser_name=parser_name, logger=utils.get_logger(__name__), **kwargs)

    def normalize(self, normalizer: typing.Union[str, typing.Callable], entry_archive=None):
        '''
        Parse the downloaded entry and run the given normalizer.
        '''
        if entry_archive is None:
            entry_archive = self.parse()

        return normalize(
            entry_archive=entry_archive, normalizer=normalizer, logger=utils.get_logger(__name__))

    def normalize_all(self, entry_archive=None):
        '''
        Parse the downloaded entry and run the whole normalizer chain.
        '''
        return normalize_all(
            entry_archive=entry_archive, logger=utils.get_logger(__name__))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any
from fastapi import FastAPI, status, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse, RedirectResponse
import traceback
import orjson

from nomad import config, utils

from .common import root_path
from .routers import users, entries, materials, auth, info, datasets, uploads, suggestions


logger = utils.get_logger(__name__)


class ORJSONResponse(JSONResponse):
    media_type = "application/json"

    def render(self, content: Any) -> bytes:
        return orjson.dumps(content, option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS)  # type: ignore


app = FastAPI(
    root_path=root_path,
    openapi_url='/openapi.json',
    docs_url='/extensions/docs',
    redoc_url='/extensions/redoc',
    swagger_ui_oauth2_redirect_url='/extensions/docs/oauth2-redirect',
    title='NOMAD API',
    version='v1, NOMAD %s@%s' % (config.meta.version, config.meta.commit),
    description=utils.strip(f'''
        Please visit the [API section of the NOMAD documentation]({config.api_url(True, 'docs/api.html')})
        for a introduction and examples.
    '''),
    default_response_class=ORJSONResponse)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


async def redirect_to_docs(req: Request):
    return RedirectResponse(f'{root_path}/extensions/docs')


# app.add_route(f'{root_path}', redirect_to_docs, include_in_schema=False)
app.add_route('/', redirect_to_docs, include_in_schema=False)


@app.exception_handler(Exception)
async def unicorn_exception_handler(request: Request, e: Exception):
    logger.error('unexpected exception in API', url=request.url, exc_info=e)
    return JSONResponse(
        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
        content={
            'detail': {
                'reason': 'Unexpected exception while handling your request',
                'exception': str(e),
                'exception_class': e.__class__.__name__,
                'exception_traceback': traceback.format_exc()
            }
        }
    )

app.include_router(info.router, prefix='/info')
app.include_router(auth.router, prefix='/auth')
app.include_router(materials.router, prefix='/materials')
app.include_router(entries.router, prefix='/entries')
app.include_router(datasets.router, prefix='/datasets')
app.include_router(uploads.router, prefix='/uploads')
app.include_router(users.router, prefix='/users')
app.include_router(suggestions.router, prefix='/suggestions')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference, JSON
)

from siestaparser.metainfo import siesta_autogenerated
from nomad.datamodel.metainfo import simulation


m_package = Package()


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_siesta_system_label = Quantity(
        type=str,
        shape=[],
        description='''
        system label specified in siesta calculation.  This determines many of the
        filenames.
        ''')

    x_siesta_system_name = Quantity(
        type=str,
        shape=[],
        description='''
        system name specified in siesta calculation.  This is purely a description.
        ''')

    x_siesta_arch = Quantity(
        type=str,
        shape=[],
        description='''
        architecture
        ''')

    x_siesta_compiler_flags = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_siesta_parallel = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_siesta_n_nodes = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')


class x_siesta_section_input(siesta_autogenerated.x_siesta_section_input):

    m_def = Section(validate=False, extends_base_section=True)

    x_siesta_xc_authors = Quantity(
        type=str,
        shape=[],
        description='''
        XC authors (really: XC functional) in siesta calculation.
        ''')

    x_siesta_typeofrun = Quantity(
        type=str,
        shape=[],
        description='''
        type of run as categorized by Siesta (optimization/md etc.)
        ''')


class BasisSet(simulation.method.BasisSet):

    m_def = Section(validate=False, extends_base_section=True)

    x_siesta_pao_basis_type = Quantity(
        type=str,
        shape=[],
        description='''
        basis set type
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_siesta_input_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')

    x_siesta_simulation_parameters = Quantity(
        type=JSON,
        shape=[],
        description='''
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Dict, Set, Iterator, Any, Optional, Union
from types import FunctionType
import urllib
import io
import json
import os
import inspect
from fastapi import Request, Query, HTTPException, status  # pylint: disable=unused-import
from pydantic import ValidationError, BaseModel  # pylint: disable=unused-import
import gzip
import lzma
from nomad.files import UploadFiles, StreamedFile, create_zipstream


def parameter_dependency_from_model(name: str, model_cls, exclude: List[str] = []):
    '''
    Takes a pydantic model class as input and creates a dependency with corresponding
    Query parameter definitions that can be used for GET
    requests.

    This will only work, if the fields defined in the input model can be turned into
    suitable query parameters. Otherwise fastapi will complain down the road.

    Arguments:
        name: Name for the dependency function.
        model_cls: A ``BaseModel`` inheriting model class as input.
    '''
    names = []
    annotations: Dict[str, type] = {}
    defaults = []
    for field_model in model_cls.__fields__.values():
        if field_model.name not in exclude:
            field_info = field_model.field_info
            names.append(field_model.name)
            annotations[field_model.name] = field_model.outer_type_
            defaults.append(Query(field_model.default, description=field_info.description))

    code = inspect.cleandoc('''
    def %s(%s):
        try:
            return %s(%s)
        except ValidationError as e:
            errors = e.errors()
            for error in errors:
                error['loc'] = ['query'] + list(error['loc'])
            raise HTTPException(422, detail=errors)

    ''' % (
        name, ', '.join(names), model_cls.__name__,
        ', '.join(['%s=%s' % (name, name) for name in names])))

    compiled = compile(code, 'string', 'exec')
    env = {model_cls.__name__: model_cls}
    env.update(**globals())
    func = FunctionType(compiled.co_consts[0], env, name)
    func.__annotations__ = annotations
    func.__defaults__ = (*defaults,)

    return func


class DownloadItem(BaseModel):
    ''' Defines an object (file or folder) for download. '''
    upload_id: str
    raw_path: str
    zip_path: str
    entry_metadata: Optional[Dict[str, Any]]


def create_download_stream_zipped(
        download_items: Union[DownloadItem, Iterator[DownloadItem]],
        upload_files: UploadFiles = None,
        re_pattern: Any = None,
        recursive: bool = False,
        create_manifest_file: bool = False,
        compress: bool = True) -> Iterator[bytes]:
    '''
    Creates a zip-file stream for downloading raw data with ``StreamingResponse``.

    Arguments:
        download_items: A DownloadItem, or iterator of DownloadItems, defining what to download.
        upload_files: The UploadFiles object, if already opened (for optimiztion).
        re_pattern: A regex object for filtering by filenames (only applicable to directories).
        recursive: if subdirectories should be included (recursively).
        create_manifest_file: if set, a manifest file is created in the root folder.
        compress: if the zip file should be compressed or not
    '''
    def streamed_files(upload_files) -> Iterator[StreamedFile]:
        manifest = []
        try:
            items: Iterator[DownloadItem] = (
                iter([download_items]) if isinstance(download_items, DownloadItem)
                else download_items)
            streamed_paths: Set[str] = set()

            for download_item in items:
                if upload_files and upload_files.upload_id != download_item.upload_id:
                    # We're switching to a new upload. Close the old.
                    upload_files.close()
                    upload_files = None
                if not upload_files:
                    # Open the requested upload.
                    upload_files = UploadFiles.get(download_item.upload_id)

                all_filtered = True
                files_found = False
                if not upload_files.raw_path_exists(download_item.raw_path):
                    pass
                elif upload_files.raw_path_is_file(download_item.raw_path):
                    # File
                    if download_item.zip_path not in streamed_paths:
                        streamed_paths.add(download_item.zip_path)
                        yield StreamedFile(
                            path=download_item.zip_path,
                            f=upload_files.raw_file(download_item.raw_path, 'rb'),
                            size=upload_files.raw_file_size(download_item.raw_path))
                else:
                    # Directory
                    for path_info in upload_files.raw_directory_list(
                            download_item.raw_path, recursive, files_only=True):
                        files_found = True
                        if not re_pattern or re_pattern.search(path_info.path):
                            all_filtered = False
                            relative_path = os.path.relpath(path_info.path, download_item.raw_path)
                            zip_path = os.path.join(download_item.zip_path, relative_path)
                            if zip_path not in streamed_paths:
                                streamed_paths.add(zip_path)
                                yield StreamedFile(
                                    path=zip_path,
                                    f=upload_files.raw_file(path_info.path, 'rb'),
                                    size=path_info.size)

                if create_manifest_file and download_item.entry_metadata:
                    if not all_filtered or not files_found:
                        manifest.append(download_item.entry_metadata)

            if create_manifest_file:
                manifest_content = json.dumps(manifest).encode()
                yield StreamedFile(path='manifest.json', f=io.BytesIO(manifest_content), size=len(manifest_content))

        finally:
            if upload_files:
                upload_files.close()

    return create_zipstream(streamed_files(upload_files), compress=compress)


def create_download_stream_raw_file(
        upload_files: UploadFiles, path: str,
        offset: int = 0, length: int = -1, decompress=False) -> Iterator[bytes]:
    '''
    Creates a file stream for downloading raw data with ``StreamingResponse``.

    Arguments:
        upload_files: the UploadFiles object, containing the file. Will be closed when done.
        path: the raw path within the upload to the desired file.
        offset: offset within the file (0 by default)
        length: number of bytes to read. -1 by default, which means the remainder of the
            file will be read.
        decompress: decompresses if the file is compressed (and of a supported type).
    '''
    raw_file: Any = upload_files.raw_file(path, 'rb')
    if decompress:
        if path.endswith('.gz'):
            raw_file = gzip.GzipFile(filename=path[:3], mode='rb', fileobj=raw_file)

        if path.endswith('.xz'):
            raw_file = lzma.open(filename=raw_file, mode='rb')

    assert offset >= 0, 'Invalid offset provided'
    assert length > 0 or length == -1, 'Invalid length provided. Should be > 0 or equal to -1.'
    if offset > 0:
        raw_file.seek(offset)

    if length > 0:
        # Read up to a certain number of bytes
        remaining = length
        while remaining:
            content = raw_file.read(remaining)
            content_length = len(content)
            remaining -= content_length
            if content_length == 0:
                break  # No more bytes
            yield content
    else:
        # Read until the end of the file.
        while True:
            content = raw_file.read(1024 * 64)
            if not content:
                break  # No more bytes
            yield content

    raw_file.close()
    upload_files.close()


def create_stream_from_string(content: str) -> io.BytesIO:
    ''' For returning strings as content using '''
    return io.BytesIO(content.encode())


def create_responses(*args):
    return {
        status_code: response
        for status_code, response in args}


def browser_download_headers(filename: str, media_type: str = 'application/octet-stream') -> Dict[str, str]:
    '''
    Creates standardized headers which tells browsers that they should download the
    data to a file with the specified filename. Note, the `media_type` should normally be
    either `application/octet-stream` or `application/zip`, using for example `application/json`
    will cause most browsers to try to open and view the file instead of downloading it.
    '''
    assert filename, 'Filename must be specified'
    filename = filename.replace('"', '\\"')
    return {
        'Content-Type': media_type,
        'Content-Disposition': f'attatchment; filename="{filename}"'}


def update_url_query_arguments(original_url: str, **kwargs) -> str:
    '''
    Takes an url, and returns a new url, obtained by updating the query arguments in the
    `original_url` as specified by the kwargs.
    '''
    scheme, netloc, path, params, query, fragment = urllib.parse.urlparse(original_url)
    query_dict = urllib.parse.parse_qs(query)
    for k, v in kwargs.items():
        if v is None:
            # Unset the value
            if k in query_dict:
                query_dict.pop(k)
        else:
            query_dict[k] = [str(v)]
    query = urllib.parse.urlencode(query_dict, doseq=True)
    return urllib.parse.urlunparse((scheme, netloc, path, params, query, fragment))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import click


from nomad.cli.cli import cli
from nomad import config


@cli.group(help='Commands that use the nomad API to do useful things')
@click.option('-n', '--url', default=config.client.url, help='The URL where nomad is running, default is "%s".' % config.client.url)
@click.option('-u', '--user', default=None, help='the user name to login, default is no login.')
@click.option('-w', '--password', default=config.client.password, help='the password used to login.')
@click.option('--token-via-api', is_flag=True, help='retrieve the access token from the api not keycloak.')
@click.option('--no-ssl-verify', help='disables SSL verificaton when talking to nomad.', is_flag=True)
@click.pass_context
def client(ctx, url: str, user: str, password: str, no_ssl_verify: bool, token_via_api: bool):
    config.client.url = url

    ctx.obj.user = user
    ctx.obj.password = password
    ctx.obj.token_via_api = token_via_api


def _create_auth(ctx):
    print(f'Used nomad is {config.client.url}')
    print(f'Used user is {ctx.obj.user}')

    from nomad.client import Auth
    if ctx.obj.user is None:
        return None

    return Auth(user=ctx.obj.user, password=ctx.obj.password, from_api=ctx.obj.token_via_api)


@client.command(help='Runs a few example operations as a test.')
@click.option(
    '--skip-parsers', is_flag=True,
    help='Skip extensive upload and parser tests.')
@click.option(
    '--skip-publish', is_flag=True,
    help='Skip publish the upload. Should not be done on an production environment.')
@click.option(
    '--skip-doi', is_flag=True,
    help='Skip assigning a doi to a dataset.')
@click.pass_context
def integrationtests(ctx, skip_parsers, skip_publish, skip_doi):
    import sys

    if ctx.obj.user is None:
        print('Authorization is required. Please provide credentials via --user, --password.')
        sys.exit(1)

    from .integrationtests import integrationtests
    integrationtests(_create_auth(ctx), skip_parsers, skip_publish, skip_doi)


@client.command(
    help='Upload files to nomad. The given path can be a single file or a directory. '
    'All .zip files in a directory will be uploaded.')
@click.argument('PATH', nargs=-1, required=True, type=click.Path(exists=True))
@click.option(
    '--upload-name',
    help='Optional name for the upload of a single file. Will be ignored on directories.')
@click.option(
    '--local-path', is_flag=True, default=False,
    help='Upload files "offline": files will not be uploaded, but processed were they are. '
    'Only works when run on the nomad host.')
@click.option(
    '--publish', is_flag=True, default=False,
    help='Automatically move upload out of the staging area after successful processing')
@click.pass_context
def upload(ctx, path, upload_name: str, local_path: bool, publish: bool):
    import os
    import sys

    from nomad.client import upload_file as client_upload_file

    auth = _create_auth(ctx)
    paths = path

    def upload_file(path, upload_name):
        result = client_upload_file(
            path, auth, upload_name=upload_name, local_path=local_path, publish=publish)
        if result is None:
            sys.exit(1)

    click.echo('uploading files from %s paths' % len(paths))
    for path in paths:
        click.echo('uploading %s' % path)
        if os.path.isfile(path):
            upload_name = upload_name if upload_name is not None else os.path.basename(path)
            upload_file(path, upload_name)

        elif os.path.isdir(path):
            for (dirpath, _, filenames) in os.walk(path):
                for filename in filenames:
                    if filename.endswith('.zip'):
                        file_path = os.path.abspath(os.path.join(dirpath, filename))
                        upload_name = os.path.basename(file_path)
                        upload_file(file_path, upload_name)

        else:
            click.echo('Unknown path type %s.' % path)


@client.command(help='Run processing locally.')
@click.argument('ENTRY_ID', nargs=1, required=True, type=str)
@click.option('--override', is_flag=True, help='Override existing local entry data.')
@click.option('--show-archive', is_flag=True, help='Print the archive data.')
@click.option('--show-metadata', is_flag=True, help='Print the extracted repo metadata.')
@click.option('--skip-normalizers', is_flag=True, help='Do not normalize.')
@click.option('--not-strict', is_flag=True, help='Also match artificial parsers.')
@click.pass_context
def local(ctx, entry_id, show_archive, show_metadata, skip_normalizers, not_strict, **kwargs):
    import sys
    import json

    from nomad.client import LocalEntryProcessing

    print('Using %s' % config.client.url)
    auth = _create_auth(ctx)

    with LocalEntryProcessing(entry_id, auth=auth, **kwargs) as local:
        print(f'Data being saved to .volumes/fs/tmp/repro_{entry_id} if not already there')
        entry_archives = local.parse(strict=not not_strict)

        for entry_archive in entry_archives:
            if not skip_normalizers:
                local.normalize_all(entry_archive=entry_archive)

            if show_archive:
                json.dump(entry_archive.m_to_dict(), sys.stdout, indent=2)

            if show_metadata:
                metadata = entry_archive.metadata
                metadata.apply_archvie_metadata(entry_archive)
                json.dump(metadata.m_to_dict(), sys.stdout, indent=4)


@client.command(
    help='Synchronizes the NOMAD database with the given external database.')
@click.argument('db_name', nargs=1, required=True)
@click.argument('root_url', nargs=1, required=True)
@click.option(
    '--outfile', default=None,
    help='File to read/write files missing in NOMAD database')
@click.option(
    '--nomadfile', default=None,
    help='File to read/write files in NOMAD database')
@click.option(
    '--dbfile', default=None,
    help='File to read/write files in given database')
@click.option(
    '--local_path', default='/nomad/fairdi/external',
    help='Directory to which the files will be downloaded')
@click.option(
    '--parallel', default=2,
    help='Number of processes to spawn to download/upload files')
@click.option(
    '--do-download', is_flag=True, default=False,
    help='Flag to automatically download downloaded files')
@click.option(
    '--do-upload', is_flag=True, default=False,
    help='Flag to automatically upload downloaded files')
@click.option(
    '--do-publish', is_flag=True, default=False,
    help='Flag to automatically publish upload')
@click.option(
    '--cleanup', is_flag=True, default=False,
    help='Flag to clean up downloaded files')
@click.pass_context
def synchdb(ctx, **kwargs):
    from nomad.cli.aflow import DbUpdater

    db = DbUpdater(auth=_create_auth(ctx), **kwargs)
    db.update()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)

from nomad.datamodel.metainfo import simulation


m_package = Package()


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_aflow_aurl = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_auid = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_data_api = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_data_source = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_n_loop = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_loop = Quantity(
        type=str,
        shape=['x_aflow_n_loop'],
        description='''
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_aflow_code = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_species_pp = Quantity(
        type=str,
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_n_dft_type = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_dft_type = Quantity(
        type=str,
        shape=['x_aflow_n_dft_type'],
        description='''
        ''')

    x_aflow_dft_type = Quantity(
        type=str,
        shape=['x_aflow_n_dft_type'],
        description='''
        ''')

    x_aflow_species_pp_version = Quantity(
        type=str,
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_species_pp_ZVAL = Quantity(
        type=np.dtype(np.int32),
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_species_pp_AUID = Quantity(
        type=str,
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_ldau_type = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_ldau_l = Quantity(
        type=np.dtype(np.int32),
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_ldau_u = Quantity(
        type=np.dtype(np.int32),
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_ldau_j = Quantity(
        type=np.dtype(np.int32),
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_valence_cell_iupac = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_valence_cell_std = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_energy_cutoff = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_delta_electronic_energy_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_delta_electronic_energy_threshold = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_kpoints_relax = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        ''')

    x_aflow_kpoints_static = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        ''')

    x_aflow_n_kpoints_bands_path = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_kpoints_bands_path = Quantity(
        type=str,
        shape=['x_aflow_n_kpoints_bands_path'],
        description='''
        ''')

    x_aflow_kpoints_bands_nkpts = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_aflow_compound = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_prototype = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_nspecies = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_natoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_natoms_orig = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_composition = Quantity(
        type=np.dtype(np.int32),
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_density = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_density_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_scintillation_attenuation_length = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_stoichiometry = Quantity(
        type=np.dtype(np.float64),
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_species = Quantity(
        type=str,
        shape=['x_aflow_nspecies'],
        description='''
        ''')

    x_aflow_geometry = Quantity(
        type=np.dtype(np.float64),
        shape=[6],
        description='''
        ''')

    x_aflow_geometry_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[6],
        description='''
        ''')

    x_aflow_volume_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_volume_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_volume_cell_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_volume_atom_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_n_sg = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_sg = Quantity(
        type=str,
        shape=['x_aflow_n_sg'],
        description='''
        ''')

    x_aflow_sg2 = Quantity(
        type=str,
        shape=['x_aflow_n_sg'],
        description='''
        ''')

    x_aflow_spacegroup_orig = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_spacegroup_relax = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_lattice_variation_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_lattice_system_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Pearson_symbol_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_relax = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_lattice_variation_relax = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_lattice_system_relax = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Pearson_symbol_relax = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_crystal_family_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_crystal_system_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_crystal_class_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_Hermann_Mauguin_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_Schoenflies_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_orbifold_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_type_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_order_orig = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_point_group_structure_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_lattice_type_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_lattice_variation_type_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_lattice_system_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_superlattice_lattice_type_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_superlattice_lattice_variation_type_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_superlattice_lattice_system_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Pearson_symbol_superlattice_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_reciprocal_geometry_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[6],
        description='''
        ''')

    x_aflow_reciprocal_volume_cell_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_reciprocal_lattice_type_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_reciprocal_lattice_variation_type_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_n_symmetries = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_Wyckoff_letters_orig = Quantity(
        type=np.dtype('U'),
        shape=['x_aflow_nspecies', 'x_aflow_n_symmetries'],
        description='''
        ''')

    x_aflow_Wyckoff_multiplicities_orig = Quantity(
        type=np.dtype(np.int32),
        shape=['x_aflow_nspecies', 'x_aflow_n_symmetries'],
        description='''
        ''')

    x_aflow_Wyckoff_site_symmetries_orig = Quantity(
        type=np.dtype('U'),
        shape=['x_aflow_nspecies', 'x_aflow_n_symmetries'],
        description='''
        ''')

    x_aflow_crystal_family = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_crystal_system = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_crystal_class = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_Hermann_Mauguin = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_Schoenflies = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_orbifold = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_point_group_order = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_point_group_structure = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_lattice_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_lattice_variation_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_lattice_lattice_system = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_superlattice_lattice_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_superlattice_lattice_variation_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Bravais_superlattice_lattice_system = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Pearson_symbol_superlattice = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_reciprocal_geometry = Quantity(
        type=np.dtype(np.float64),
        shape=[6],
        description='''
        ''')

    x_aflow_reciprocal_volume_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_reciprocal_lattice_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_reciprocal_lattice_variation_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_Wyckoff_letters = Quantity(
        type=np.dtype('U'),
        shape=['x_aflow_nspecies', 1],
        description='''
        ''')

    x_aflow_Wyckoff_multiplicities = Quantity(
        type=np.dtype(np.int32),
        shape=['x_aflow_nspecies', 1],
        description='''
        ''')

    x_aflow_Wyckoff_site_symmetries = Quantity(
        type=np.dtype('U'),
        shape=['x_aflow_nspecies', 1],
        description='''
        ''')

    x_aflow_prototype_label_orig = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_prototype_params_list_orig = Quantity(
        type=str,
        shape=[3],
        description='''
        ''')

    x_aflow_prototype_params_values_orig = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        ''')

    x_aflow_prototype_label_relax = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_prototype_params_list_relax = Quantity(
        type=str,
        shape=[3],
        description='''
        ''')

    x_aflow_prototype_params_values_relax = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        ''')


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_aflow_stress_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[9],
        description='''
        ''')

    x_aflow_pressure_residual = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_Pulay_stress = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_Egap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_Egap_fit = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_Egap_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_enthalpy_formation_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        ''')

    x_aflow_entropic_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_PV = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_spin_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_spinD = Quantity(
        type=np.dtype(np.float64),
        shape=['x_aflow_natoms'],
        description='''
        ''')

    x_aflow_spinF = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_calculation_memory = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_calculation_cores = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_nbondxx = Quantity(
        type=np.dtype(np.float64),
        shape=[6],
        description='''
        ''')

    x_aflow_agl_thermal_conductivity_300K = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_debye = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_acoustic_debye = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_gruneisen = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_heat_capacity_Cv_300K = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_heat_capacity_Cp_300K = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_thermal_expansion_300K = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_bulk_modulus_static_300K = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_bulk_modulus_isothermal_300K = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_poisson_ratio_source = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_agl_vibrational_free_energy_300K_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_vibrational_free_energy_300K_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_vibrational_entropy_300K_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_agl_vibrational_entropy_300K_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_poisson_ratio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_bulk_modulus_voigt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_bulk_modulus_reuss = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_shear_modulus_voigt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_shear_modulus_reuss = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_bulk_modulus_vrh = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_shear_modulus_vrh = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_elastic_anisotropy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_youngs_modulus_vrh = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_speed_sound_transverse = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_speed_sound_longitudinal = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_speed_sound_average = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_pughs_modulus_ratio = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_debye_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_applied_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_average_external_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_aflow_ael_stiffness_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6],
        description='''
        ''')

    x_aflow_ael_compliance_tensor = Quantity(
        type=np.dtype(np.float64),
        shape=[6, 6],
        description='''
        ''')

    x_aflow_bader_net_charges = Quantity(
        type=np.dtype(np.float64),
        shape=['x_aflow_natoms'],
        description='''
        ''')

    x_aflow_bader_atomic_volumes = Quantity(
        type=np.dtype(np.float64),
        shape=['x_aflow_natoms'],
        description='''
        ''')

    x_aflow_n_files = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_files = Quantity(
        type=str,
        shape=['x_aflow_n_files'],
        description='''
        ''')

    x_aflow_node_CPU_Model = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_node_CPU_Cores = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_node_CPU_MHz = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_node_RAM_GB = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_aflow_catalog = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_aflowlib_version = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_aflow_aflowlib_date = Quantity(
        type=np.dtype(np.float64),
        shape=['x_aflow_nspecies'],
        description='''
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Tuple, Dict
from urllib.parse import urljoin, urlsplit, urlunsplit
import re
import json

from nomad import utils, config
from nomad.metainfo import Context as MetainfoContext, MSection, Quantity, MetainfoReferenceError
from nomad.datamodel import EntryArchive


class Context(MetainfoContext):
    '''
    The nomad implementation of a metainfo context.
    '''

    def __init__(self, installation_url: str = None):
        self.installation_url = installation_url
        if self.installation_url is None:
            self.installation_url = config.api_url(api='api/v1')

        self.archives: Dict[str, MSection] = {}
        self.urls: Dict[MSection, str] = {}

    @property
    def upload_id(self):
        return None

    def _get_ids(self, root: MSection) -> Tuple[str, str]:
        if not isinstance(root, EntryArchive):
            return None, None

        if root.metadata:
            return root.metadata.upload_id, root.metadata.entry_id

        return None, None

    def create_reference(self, section: MSection, quantity_def: Quantity, value: MSection) -> str:
        fragment = value.m_path()

        source_root: MSection = section.m_root()
        target_root: MSection = value.m_root()

        if source_root == target_root:
            return f'#{fragment}'

        if target_root in self.urls:
            return f'{self.urls[target_root]}#{fragment}'

        upload_id, entry_id = self._get_ids(target_root)
        assert entry_id is not None, 'Only archives with entry_id can be referenced'
        assert upload_id is not None, 'Only archives with upload_id can be referenced'
        assert source_root.m_context == self, 'Can only create references from archives with this context'
        installation_url = getattr(target_root.m_context, 'installation_url', self.installation_url)

        if installation_url != self.installation_url:
            return urljoin(installation_url, f'/api/v1/uploads/{upload_id}/archive/{entry_id}#{fragment}')

        source_upload_id, source_entry_id = self._get_ids(source_root)

        if entry_id == source_entry_id:
            return f'#{fragment}'

        if upload_id == source_upload_id:
            return f'../upload/archive/{entry_id}#{fragment}'

        return f'../uploads/{upload_id}/archive/{entry_id}#{fragment}'

    def _normalize_fragment(self, fragment):
        if fragment and fragment != '' and not fragment.startswith('/'):
            fragment = f'/{fragment}'
        return fragment

    def normalize_reference(self, source: MSection, url: str) -> str:
        ''' Replace mainfile references with entry-based references. '''
        url_parts = urlsplit(url)
        path = url_parts.path
        match = re.search(r'/archive/mainfile/(.*)$', path)
        if not match:
            return urlunsplit(url_parts[0:4] + (self._normalize_fragment(url_parts.fragment),))

        mainfile = match.group(1)
        upload_id = self.upload_id
        if upload_id is None:
            root_section: MSection = source.m_root()
            upload_id = root_section.metadata.upload_id
        assert upload_id is not None, 'Only archives with upload_id can be referenced'
        entry_id = utils.generate_entry_id(upload_id, mainfile)
        path = path.replace(f'/archive/mainfile/{mainfile}', f'/archive/{entry_id}')
        return urlunsplit((
            url_parts.scheme,
            url_parts.netloc,
            path,
            url_parts.query,
            self._normalize_fragment(url_parts.fragment),))

    def load_archive(self, entry_id: str, upload_id: str, installation_url: str) -> EntryArchive:
        ''' Loads the archive for the given identification. '''
        raise NotImplementedError()

    def load_raw_file(self, path: str, upload_id: str, installation_url: str) -> MSection:
        ''' Loads a raw file based on the given upload and path. Interprets is as metainfo data. '''
        raise NotImplementedError()

    def _parse_url(self, url: str) -> Tuple[str, str, str, str]:
        installation_url = self.installation_url
        upload_id = self.upload_id

        url_match = re.match(r'^../upload/(archive|raw)/([^\?]+)$', url)
        if url_match:
            kind = url_match.group(1)
            path = url_match.group(2)
            return installation_url, upload_id, kind, path

        url_match = re.match(r'^../uploads/([A-Za-z0-9_]+)/(archive|raw)/([^\?]+)$', url)
        if url_match:
            upload_id = url_match.group(1)
            kind = url_match.group(2)
            path = url_match.group(3)
            return installation_url, upload_id, kind, path

        url_match = re.search(r'/api/v1/uploads/([A-Za-z0-9_]+)/(archive|raw)/([^\?]+)$', url)
        if url_match:
            installation_url = url.replace(url_match.group(0), '')
            upload_id = url_match.group(1)
            kind = url_match.group(2)
            path = url_match.group(3)
            return installation_url, upload_id, kind, path

        raise MetainfoReferenceError(f'the url {url} is not a valid metainfo reference')

    def load_url(self, url: str) -> MSection:
        installation_url, upload_id, kind, path = self._parse_url(url)
        if kind == 'archive':
            if path.startswith('mainfile/'):
                entry_id = utils.generate_entry_id(upload_id, path.replace('mainfile/', ''))
            elif '/' in path:
                raise MetainfoReferenceError(f'the url {url} is not a valid metainfo reference')
            else:
                entry_id = path

            return self.load_archive(entry_id, upload_id, installation_url)

        if kind == 'raw':
            return self.load_raw_file(path, upload_id, installation_url)

        raise MetainfoReferenceError(f'the url {url} is not a valid metainfo reference')

    def resolve_archive_url(self, url: str) -> MSection:
        if url not in self.archives:
            archive = self.load_url(url)
            self.archives[url] = archive
            self.urls[archive] = url

        return self.archives[url]


class ServerContext(Context):
    def __init__(self, upload=None):
        super().__init__()
        self.upload = upload

    @property
    def upload_files(self):
        if self.upload:
            return self.upload.upload_files

    @property
    def upload_id(self):
        if self.upload:
            return self.upload.upload_id

    def _get_upload_files(self, upload_id: str, installation_url: str):
        if installation_url and self.installation_url != installation_url:
            raise NotImplementedError()

        if self.upload_files and self.upload_files.upload_id == upload_id:
            return self.upload_files
        else:
            # delayed import, context is part of datamodel which should be available in
            # base install, files however requires [insfrastructure].
            # TODO move server context to some infrastructure package!
            from nomad import files
            return files.UploadFiles.get(upload_id)

    def load_archive(self, entry_id: str, upload_id: str, installation_url: str) -> EntryArchive:
        upload_files = self._get_upload_files(upload_id, installation_url)

        try:
            archive_dict = upload_files.read_archive(entry_id)[entry_id].to_dict()
        except KeyError:
            raise MetainfoReferenceError(f'archive does not exist {entry_id}')

        return EntryArchive.m_from_dict(archive_dict, m_context=self)

    def load_raw_file(self, path: str, upload_id: str, installation_url: str) -> EntryArchive:
        upload_files = self._get_upload_files(upload_id, installation_url)

        try:
            with upload_files.raw_file(path, 'rt') as f:
                archive_data = json.load(f)
        except Exception:
            raise MetainfoReferenceError(f'Could not load {path}.')

        if 'm_def' not in archive_data:
            return EntryArchive.m_from_dict(archive_data, m_context=self)
        else:
            return MSection.from_dict(archive_data, m_context=self)

    def raw_file(self, *args, **kwargs):
        return self.upload_files.raw_file(*args, **kwargs)


class ClientContext(Context):
    def load_archive(self, entry_id: str, upload_id: str, installation_url: str) -> EntryArchive:
        raise NotImplementedError()

    def load_raw_file(self, path: str, upload_id: str, installation_url: str) -> MSection:
        raise NotImplementedError()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Dict
import numpy as np
import re
import ase.data
import ase.formula
from string import ascii_uppercase
import pint.quantity
from collections import OrderedDict

from nomad.normalizing.normalizer import SystemBasedNormalizer
from nomad.units import ureg
from nomad.datamodel import OptimadeEntry, Species, EntryMetadata
from nomad.datamodel.metainfo.simulation.system import Atoms, System


species_re = re.compile(r'^([A-Z][a-z]?)(\d*)$')


# TODO this should be the default and not necessary
def transform_to_v1(entry: EntryMetadata) -> EntryMetadata:
    '''
    Transformation function to use during re-indexing of entries with outdated optimade
    format. Fixes formulas and periodic dimensions, removed entries with X in formula.
    '''
    optimade = entry.optimade
    if optimade is None:
        return entry

    if optimade.chemical_formula_reduced is None or 'X' in optimade.chemical_formula_reduced:
        entry.m_remove_sub_section(EntryMetadata.optimade, -1)
        return entry

    optimade.chemical_formula_reduced = optimade_chemical_formula_reduced(optimade.chemical_formula_reduced)
    optimade.chemical_formula_anonymous = optimade_chemical_formula_anonymous(optimade.chemical_formula_reduced)
    optimade.chemical_formula_hill = optimade_chemical_formula_hill(optimade.chemical_formula_hill)
    optimade.chemical_formula_descriptive = optimade.chemical_formula_hill
    dimension_types = optimade.dimension_types
    if dimension_types is None:
        optimade.dimension_types = [0, 0, 0]
    elif isinstance(dimension_types, int):
        optimade.dimension_types = [1] * dimension_types + [0] * (3 - dimension_types)

    return entry


def optimade_chemical_formula_reduced(formula: str):
    if formula is None:
        return formula

    try:
        ase_formula = ase.formula.Formula(formula).count()
        result_formula = ''
        for element in sorted(ase_formula.keys()):
            result_formula += element
            element_count = ase_formula[element]
            if element_count > 1:
                result_formula += str(element_count)

        return result_formula
    except Exception:
        return formula


def optimade_chemical_formula_anonymous(formula: str):
    if formula is None:
        return formula

    try:
        ase_formula = ase.formula.Formula(formula).count()
        result_formula = ''
        for index, element_count in enumerate(reversed(sorted(ase_formula.values()))):
            result_formula += ascii_uppercase[index]
            if element_count > 1:
                result_formula += str(element_count)

        return result_formula
    except Exception:
        return formula


def optimade_chemical_formula_hill(formula: str):
    if formula is None:
        return formula

    try:
        ase_formula = ase.formula.Formula(formula).count()
        result: Dict[str, int] = OrderedDict()
        if 'C' in ase_formula:
            for symbol in 'CH':
                if symbol in ase_formula:
                    result[symbol] = ase_formula.pop(symbol)
        for symbol, n in sorted(ase_formula.items()):
            result[symbol] = n
        return ''.join([
            symbol + (str(n) if n > 1 else '')
            for symbol, n in result.items()])
    except Exception:
        return formula


class OptimadeNormalizer(SystemBasedNormalizer):

    '''
    This normalizer performs all produces a section all data necessary for the Optimade API.
    It assumes that the :class:`SystemNormalizer` was run before.
    '''
    def __init__(self, archive):
        super().__init__(archive, only_representatives=True)

    def add_optimade_data(self, index) -> OptimadeEntry:
        '''
        The 'main' method of this :class:`SystemBasedNormalizer`.
        Normalizes the section with the given `index`.
        Normalizes geometry, classifies, system_type, and runs symmetry analysis.
        '''
        if self.entry_archive.metadata is None:
            self.entry_archive.m_create(EntryMetadata)
        optimade = self.entry_archive.metadata.m_create(OptimadeEntry)

        def get_value(quantity_def, default: Any = None, numpy: bool = False, unit=None, source: Any = None) -> Any:
            try:
                source = source if source is not None else self.section_run.system[-1]
                value = source.m_get(quantity_def)
                if value is None:
                    return
                if type(value) == np.ndarray and not numpy:
                    return value.tolist()
                if isinstance(value, list) and numpy:
                    return np.array(value)

                if numpy and unit is not None:
                    if isinstance(value, pint.quantity._Quantity):
                        value = value.to(unit)
                    elif value is not None:
                        value = value * unit

                return value
            except KeyError:
                return default

        from nomad.normalizing.system import normalized_atom_labels

        system = self.section_run.system[-1] if self.section_run.system else None
        if system is None:
            return optimade

        nomad_species = get_value(Atoms.labels, source=system.atoms)
        nomad_species = [] if nomad_species is None else nomad_species

        # elements
        atoms = normalized_atom_labels(nomad_species)
        atom_count = len(atoms)
        atom_counts: Dict[str, int] = {}
        for atom in atoms:
            current = atom_counts.setdefault(atom, 0)
            current += 1
            atom_counts[atom] = current

        optimade.elements = list(set(atoms))
        optimade.elements.sort()
        optimade.nelements = len(optimade.elements)
        optimade.elements_ratios = [
            atom_counts[element] / atom_count
            for element in optimade.elements]

        # formulas
        optimade.chemical_formula_reduced = optimade_chemical_formula_reduced(
            get_value(System.chemical_composition_reduced, source=system))
        optimade.chemical_formula_hill = optimade_chemical_formula_hill(
            get_value(System.chemical_composition_hill, source=system))
        optimade.chemical_formula_descriptive = optimade.chemical_formula_hill
        optimade.chemical_formula_anonymous = optimade_chemical_formula_anonymous(
            optimade.chemical_formula_reduced)

        # sites
        optimade.nsites = len(nomad_species)
        optimade.species_at_sites = nomad_species
        optimade.lattice_vectors = get_value(Atoms.lattice_vectors, numpy=True, unit=ureg.m, source=system.atoms)
        optimade.cartesian_site_positions = get_value(Atoms.positions, numpy=True, unit=ureg.m, source=system.atoms)
        pbc = get_value(Atoms.periodic, source=system.atoms)
        if pbc is not None:
            optimade.dimension_types = [1 if value else 0 for value in pbc]

        # species
        for species_label in set(nomad_species):
            match = re.match(species_re, species_label)

            element_label = match.group(1) if match else species_label

            species = optimade.m_create(Species)
            species.name = species_label
            if element_label in ase.data.chemical_symbols:
                chemical_symbol = element_label
            else:
                chemical_symbol = 'x'
            species.chemical_symbols = [chemical_symbol]
            species.concentration = [1.0]

        optimade.structure_features = []

        return optimade

    def normalize_system(self, system, is_representative):
        if not is_representative:
            return False

        try:
            self.add_optimade_data(system.m_parent_index)
            return True

        except Exception as e:
            self.logger.warn('could not acquire optimade data', exc_info=e)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class x_wien2k_header(MSection):
    '''
    header (labels) of wien2k.
    '''

    m_def = Section(validate=False)

    x_wien2k_release_date = Quantity(
        type=str,
        shape=[],
        description='''
        Release date of wien2k.
        ''')

    x_wien2k_version = Quantity(
        type=str,
        shape=[],
        description='''
        Version of WIEN2k.
        ''')


class x_wien2k_section_XC(MSection):
    '''
    exchange-correlation potential, in in0
    '''

    m_def = Section(validate=False)

    x_wien2k_indxc = Quantity(
        type=str,
        shape=[],
        description='''
        exchange-correlation potential, in in0
        ''')


class x_wien2k_section_equiv_atoms(MSection):
    '''
    section containing a class of equivalent atoms
    '''

    m_def = Section(validate=False)

    x_wien2k_atom_pos_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        position of atom x in internal units
        ''')

    x_wien2k_atom_pos_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        position of atom y in internal units
        ''')

    x_wien2k_atom_pos_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        position of atom z  in internal units
        ''')

    x_wien2k_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        name of atom, labelling non-equvalent atoms
        ''')

    x_wien2k_NPT = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of radial mesh points
        ''')

    x_wien2k_RMT = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        atomic sphere radius (muffin-tin radius)
        ''')

    x_wien2k_R0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        first radial mesh point
        ''')

    x_wien2k_atomic_number_Z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        atomic number Z
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_wien2k_header = SubSection(
        sub_section=SectionProxy('x_wien2k_header'),
        repeats=True)


class ScfIteration(simulation.calculation.ScfIteration):

    m_def = Section(validate=False, extends_base_section=True)

    x_wien2k_iteration_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        scf iteration number
        ''')

    x_wien2k_nr_of_independent_atoms = Quantity(
        type=int,
        description='''
        number of independent atoms in the cell
        ''')

    x_wien2k_potential_option = Quantity(
        type=str,
        shape=[],
        description='''
        exchange correlation potential option
        ''')

    x_wien2k_system_name = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system
        ''')

    x_wien2k_total_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        total number of atoms in the cell
        ''')

    x_wien2k_lattice_const_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lattice parameter a in this calculation
        ''')

    x_wien2k_lattice_const_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lattice parameter b in this calculation
        ''')

    x_wien2k_lattice_const_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lattice parameter c in this calculation
        ''')

    x_wien2k_unit_cell_volume_bohr3 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='bohr ** 3',
        description='''
        unit cell volume
        ''')

    x_wien2k_spinpolarization = Quantity(
        type=str,
        description='''
        spinpolarization treatment
        ''')

    x_wien2k_noe = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of electrons
        ''')

    x_wien2k_nr_kpts = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of k-points
        ''')

    x_wien2k_cutoff = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Potential and charge cut-off, Ry**.5
        ''')

    x_wien2k_ene_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        energy gap in Ry
        ''')

    x_wien2k_ene_gap_eV = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='electron_volt',
        description='''
        energy gap in eV
        ''')

    x_wien2k_matrix_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        matrix size
        ''')

    x_wien2k_rkm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        rkm
        ''')

    x_wien2k_LOs = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        LOs
        ''')

    x_wien2k_mmtot = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total magnetic moment in cell
        ''')

    x_wien2k_mmint = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        magnetic moment in the interstital region
        ''')

    x_wien2k_mmi = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        magnetic moment inside the sphere
        ''')

    x_wien2k_mmi001 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        magnetic moment inside the sphere
        ''')

    x_wien2k_charge_distance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        charge distance between last 2 iterations
        ''')

    x_wien2k_for_abs = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): |F|
        ''')

    x_wien2k_for = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms', 3],
        description='''
        forces on inequivalent atoms in mRy/bohr (in the local (for each atom) cartesian coordinate
        system)
        ''')

    x_wien2k_for_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): Fx
        ''')

    x_wien2k_for_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): Fy
        ''')

    x_wien2k_for_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        force on atom xx in mRy/bohr (in the local (for each atom) cartesian coordinate
        system): Fz
        ''')

    x_wien2k_for_gl = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms', 3],
        description='''
        force on inequivalent atoms (in the global coordinate system of the unit cell (in
        the same way as the atomic positions are specified))
        ''')

    x_wien2k_for_x_gl = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        force on inequivalent atom xx (in the global coordinate system of the unit cell (in
        the same way as the atomic positions are specified)): Fx
        ''')

    x_wien2k_for_y_gl = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        force on inequivalent atom xx in (in the global coordinate system of the unit cell (in
        the same way as the atomic positions are specified)): Fy
        ''')

    x_wien2k_for_z_gl = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        force on inequivalent atom xx in (in the global coordinate system of the unit cell (in
        the same way as the atomic positions are specified)): Fz
        ''')

    x_wien2k_atom_nr = Quantity(
        type=str,
        shape=[],
        description='''
        number of atom, labelling atoms
        ''')

    x_wien2k_atom_mult = Quantity(
        type=np.dtype(np.int32),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        atom multiplicity
        ''')

    x_wien2k_sphere_nr = Quantity(
        type=str,
        shape=[],
        description='''
        number of sphere, labelling spheres
        ''')

    x_wien2k_tot_diff_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        total difference charge density for atom xx between last 2 iterations
        ''')

    x_wien2k_tot_int_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total interstitial charge (mixed after MIXER)
        ''')

    x_wien2k_tot_charge_in_sphere = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        total charge in sphere xx (mixed after MIXER)
        ''')

    x_wien2k_tot_int_charge_nm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total interstitial charge (new (not mixed) from LAPW2+LCORE
        ''')

    x_wien2k_tot_charge_in_sphere_nm = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        total charge in sphere xx (new (not mixed) from LAPW2+LCORE
        ''')

    x_wien2k_tot_val_charge_cell = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total valence charge inside unit cell
        ''')

    x_wien2k_tot_val_charge_sphere = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        total valence charge in sphere xx
        ''')

    x_wien2k_density_at_nucleus_valence = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        density for atom xx at the nucleus (first radial mesh point); valence
        ''')

    x_wien2k_density_at_nucleus_semicore = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        density for atom xx at the nucleus (first radial mesh point); semi-core
        ''')

    x_wien2k_density_at_nucleus_core = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        density for atom xx at the nucleus (first radial mesh point); core
        ''')

    x_wien2k_density_at_nucleus_tot = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        density for atom xx at the nucleus (first radial mesh point); total
        ''')

    x_wien2k_nuclear_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        nuclear and electronic charge; normalization check of electronic charge densities.
        If a significant amount of electrons is missing, one might have core states, whose
        charge density is not completely confined within the respective atomic sphere. In
        such a case the corresponding states should be treated as band states (using LOs).
        ''')

    x_wien2k_electronic_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['x_wien2k_nr_of_independent_atoms'],
        description='''
        nuclear and electronic charge; normalization check of electronic charge densities.
        If a significant amount of electrons is missing, one might have core states, whose
        charge density is not completely confined within the respective atomic sphere. In
        such a case the corresponding states should be treated as band states (using LOs).
        ''')

    x_wien2k_necnr = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of the nec test, labelling nec
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_wien2k_nonequiv_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        number of inequivalent atoms in the unit cell
        ''')

    x_wien2k_system_nameIn = Quantity(
        type=str,
        shape=[],
        description='''
        user given name for this system given in the struct file
        ''')

    x_wien2k_calc_mode = Quantity(
        type=str,
        shape=[],
        description='''
        relativistic or nonrelativistic calculation mode
        ''')

    x_wien2k_unit_cell_param_a = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        unit cell parameters - a
        ''')

    x_wien2k_unit_cell_param_b = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        unit cell parameters - b
        ''')

    x_wien2k_unit_cell_param_c = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        unit cell parameters - c
        ''')

    x_wien2k_angle_between_unit_axis_alfa = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell parameters - alfa
        ''')

    x_wien2k_angle_between_unit_axis_beta = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell parameters - beta
        ''')

    x_wien2k_angle_between_unit_axis_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        unit cell parameters - gamma
        ''')

    x_wien2k_section_equiv_atoms = SubSection(
        sub_section=SectionProxy('x_wien2k_section_equiv_atoms'),
        repeats=True)


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_wien2k_switch = Quantity(
        type=str,
        shape=[],
        description='''
        switch in in0 between TOT, KXC, POT, MULT, COUL, EXCH
        ''')

    x_wien2k_ifft = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        FFT-mesh parameters in x direction for the calculation of the XC-potential in the
        interstitial region, in in0
        ''')

    x_wien2k_ifft_x = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FFT-mesh parameters in x direction for the calculation of the XC-potential in the
        interstitial region, in in0
        ''')

    x_wien2k_ifft_y = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FFT-mesh parameters in y direction for the calculation of the XC-potential in the
        interstitial region, in in0
        ''')

    x_wien2k_ifft_z = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        FFT-mesh parameters in z direction for the calculation of the XC-potential in the
        interstitial region, in in0
        ''')

    x_wien2k_ifft_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Multiplicative factor to the IFFT grid, in in0
        ''')

    x_wien2k_iprint = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        optional print switch, in in0
        ''')

    x_wien2k_wf_switch = Quantity(
        type=str,
        shape=[],
        description='''
        wave function switch between WFFIL, SUPWF, WPPRI
        ''')

    x_wien2k_rkmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        RmtKmax - determines matrix size (convergence), where Kmax is the plane wave cut-
        off, Rmt is the smallest of all atomic sphere radii
        ''')

    x_wien2k_in2_switch = Quantity(
        type=str,
        shape=[],
        description='''
        switch, in in2 between (TOT,FOR,QTL,EFG,ALM,CLM,FERMI)
        ''')

    x_wien2k_in2_emin = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        lower energy cut-off for defining the range of occupied states; in in2
        ''')

    x_wien2k_in2_ne = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        number of electrons (per unit cell) in given energy range in in2
        ''')

    x_wien2k_in2_espermin = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        LAPW2 tries to find the .mean. energies for each l channel, for both the valence
        and the semicore states. To define .valence. and .semicore. it starts at (EF -
        .esepermin.) and searches for a .gap. with a width of at least .eseper0. and
        defines this as separation energy of valence and semicore; in in2
        ''')

    x_wien2k_in2_esper0 = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        minimum gap width; in in2
        ''')

    x_wien2k_smearing_kind = Quantity(
        type=str,
        shape=[],
        description='''
        determines how EF is determined; in in2
        ''')

    x_wien2k_in2_gmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        max. G (magnitude of largest vector) in charge density Fourier expansion; in in2
        ''')

    x_wien2k_section_XC = SubSection(
        sub_section=SectionProxy('x_wien2k_section_XC'),
        repeats=True)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Union, List, Iterable, Dict
import csv

from nomad.datamodel.datamodel import EntryArchive, EntryData
from nomad.metainfo import Section, Quantity, Package, Reference, SectionProxy

from .parser import MatchingParser

# We define a simple base schema for tabular data. The parser will then generate more
# specialized sections based on the table headers. These specialized defintions will use
# this as base section definition.
# TODO maybe this should be moved to nomad.datamodel.metainfo
m_package = Package()


class TableRow(EntryData):
    ''' Represents the data in one row of a table. '''
    table_ref = Quantity(
        type=Reference(SectionProxy('Table')),
        description='A reference to the table that this row is contained in.')


class Table(EntryData):
    ''' Represents a table with many rows. '''
    row_refs = Quantity(
        type=Reference(TableRow.m_def), shape=['*'],
        description='References that connect to each row. Each row is stored in it individual entry.')


m_package.__init_metainfo__()


class TabularDataParser(MatchingParser):
    # TODO this is super simple and needs extension. Currently parses files like:
    #    header_0,header_1
    #    0_0,0_1
    #    1_0,1_1
    # TODO also extend tests/parsing/test_tabular
    def __init__(self) -> None:
        super().__init__(
            name='parser/tabular', code_name='tabular data',
            mainfile_name_re=r'.*\.csv$')

    def _read_cvs(self, filename: str) -> List[List[str]]:
        with open(filename, newline='') as csvfile:
            reader = csv.reader(csvfile, delimiter=',', quotechar='#')
            return [row for row in reader]

    def is_mainfile(
        self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
        compression: str = None
    ) -> Union[bool, Iterable[str]]:
        # We use the main file regex capabilities of the superclass to check if this is a
        # .csv file
        is_tabular = super().is_mainfile(filename, mime, buffer, decoded_buffer, compression)
        if not is_tabular:
            return False

        try:
            rows = self._read_cvs(filename)
        except Exception:
            # If this cannot be parsed as a .csv file, we don't match with this file
            return False

        return [str(item) for item in range(1, len(rows))]

    def parse(
        self, mainfile: str, archive: EntryArchive, logger=None,
        child_archives: Dict[str, EntryArchive] = None
    ):
        rows = self._read_cvs(mainfile)
        header = rows[0]
        rows = rows[1:]

        # Create a schema from the header and store it in the main archive.
        # We are creating a specialized TableRow section that has a quantity
        # for each column.
        # TODO do something smart with the headers to create more complex schemas.
        # e.g. parse headers like 'sec1/sec2/quantity_name?unit=m&type=float'
        table_row_def = Section(name='TableRow', base_sections=[TableRow.m_def])
        table_row_def.quantities = [
            Quantity(name=name, type=str) for name in header]
        archive.definitions = Package(section_definitions=[table_row_def])
        MyTableRow = table_row_def.section_cls

        # Create a Table as the main archive contents. This can hold references to the
        # rows.
        table = Table()
        archive.data = table
        table_rows = []

        # Create the data for each row by instantiating the generated schema and store it
        # in the child archives. Use references to connect with the Table section in the
        # main archive.
        for index, row in enumerate(rows):
            key = str(index + 1)
            archive = child_archives[key]
            archive.data = MyTableRow(table_ref=table, **dict(zip(header, row)))  # type: ignore
            table_rows.append(archive.data)
        table.row_refs = table_rows
